#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import numpy as NPY
import math
import sys
import random
import vtk

import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils as CGU

from qtpy.QtCore import Qt, Slot
from qtpy.QtWidgets import QColorDialog
from qtpy.QtGui import (QColor, QIcon, QPixmap, QScreen)

from CGNS.NAV.Q7VTKWindow import Ui_Q7VTKWindow
from CGNS.NAV.wfingerprint import Q7Window, Q7FingerPrint
from CGNS.NAV.mparser import Mesh
from CGNS.NAV.mparser import SIZE_PATTERN, LIST_PATTERN
from CGNS.NAV.wfile import Q7File

import CGNS.NAV.wmessages as MSG

vers = vtk.vtkVersion()
VTK_VERSION_MINOR = vers.GetVTKMinorVersion()
VTK_VERSION_MAJOR = vers.GetVTKMajorVersion()


# ----------------------------------------------------------------------------
class wVTKContext(object):
    def __init__(self, cm):
        (self.vx, self.vy, self.vz) = cm.GetViewUp()
        (self.cx, self.cy, self.cz) = cm.GetFocalPoint()
        (self.px, self.py, self.pz) = cm.GetPosition()
        (self.sx, self.sy, self.sz) = (0., 0., 0.)
        (self.ox, self.oy, self.oz) = (0., 0., 0.)

    def setViewUp(self, vx, vy, vz):
        (self.vx, self.vy, self.vz) = (vx, vy, vz)

    def setFocalPoint(self, cx, cy, cz):
        (self.cx, self.cy, self.cz) = (cx, cy, cz)

    def setPosition(self, px, py, pz):
        (self.px, self.py, self.pz) = (px, py, pz)

    def dump(self):
        return """
(self.vx,self.vy,self.vz)=%s
(self.cx,self.cy,self.cz)=%s
(self.px,self.py,self.pz)=%s
(self.sx,self.sy,self.sz)=%s
(self.ox,self.oy,self.oz)=%s""" % (
            (self.vx, self.vy, self.vz),
            (self.cx, self.cy, self.cz),
            (self.px, self.py, self.pz),
            (self.sx, self.sy, self.sz),
            (self.ox, self.oy, self.oz))


# -----------------------------------------------------------------
class Q7VTK(Q7Window, Ui_Q7VTKWindow):
    def __init__(self, control, parent, node, fgprintindex, tmodel, zlist):
        if not zlist:
            pth = '/'
        else:
            pth = '<partial>'
        self._vtkstatus = False
        self._control = control
        self._fgindex = fgprintindex
        self._tree = Q7FingerPrint.getByIndex(self._fgindex).tree
        self._parent = parent
        self._master = parent
        self._vtk = None
        self._lastSelectedType = None
        if VTK_VERSION_MAJOR < 8:
            MSG.wError(self._control,
                       501, "Your VTK version is lower than v8.0, you should consider upgrading", "Oups...")
            raise RuntimeError("VTK version is too old, please upgrade")
        # FIXME: options are required for parse, then wCGNSTreeParse should
        # be called AFTER the init (that reads options)
        # if the parse fails (no data, bad data...) the window has to be deleted
        Q7Window.__init__(self, Q7Window.VIEW_VTK, control, pth, self._fgindex)
        if not self.wCGNSTreeParse(self._tree, zlist):
            return
        self._vtkstatus = True
        self._xmin = self._ymin = self._zmin = self._xmax = self._ymax = self._zmax = 0.0
        self._epix = QIcon(QPixmap(":/images/icons/empty.png"))
        self._spix = QIcon(QPixmap(":/images/icons/selected.png"))
        self._npix = QIcon(QPixmap(":/images/icons/unselected.png"))
        self._hpix = QIcon(QPixmap(":/images/icons/hidden.png"))
        self.lut = None
        self.grid_dims = None
        self.mincolor = None
        self.maxcolor = None
        self.scalarbar = None
        self.scalarbarwidget = None
        self._selected = []
        self._picked = None
        self._hidden = []
        self.actorpt = None
        self.wact = None
        self._cacheActor = {}
        self._tmodel = tmodel
        self.color = (0, 0, 0)
        self.selectionColor = (1, 0, 0)
        self.OutlineActor = None
        self._currentactor = None
        self.PickedRenderer = None
        self.PropPicked = 0
        self.controlKey = 0
        self.highlightProp = None
        self.maxColorMap = None
        self._camera = {}
        self._blackonwhite = False
        self.widgets = False
        self.resetSpinBox()
        self.display.Initialize()
        self.display.Start()
        self.display.show()
        self._vtktree = self.wCGNSTreeActors(self._tree, zlist)
        self.bX.clicked.connect(self.b_xaxis)
        self.bY.clicked.connect(self.b_yaxis)
        self.bZ.clicked.connect(self.b_zaxis)
        #      self.bReverse.clicked.connect(self.reverseSelection)
        self.bSaveVTK.clicked.connect(self.b_saveVTK)
        self.bSuffleColors.clicked.connect(self.b_shufflecolors)
        self.bBlackColor.clicked.connect(self.b_blackandwhite)
        self.bAddView.clicked.connect(self.b_saveview)
        self.bRemoveView.clicked.connect(self.b_delview)
        self.bNext.clicked.connect(self.b_next)
        self.bPrevious.clicked.connect(self.b_prev)
        self.bReset.clicked.connect(self.b_reset)
        self.bUpdateFromTree.clicked.connect(self.b_update_from)
        self.bUpdateFromVTK.clicked.connect(self.b_update_to)
        self.bScreenShot.clicked.connect(self.screenshot)
        self.selectable.clicked.connect(self.setInteractive)
        self.ColorMapMin = QColorDialog(self)
        self.ColorMapMax = QColorDialog(self)
        self.cCurrentPath.setParent(self)
        self.ColorMapMin.colorSelected[QColor].connect(self.getColorMapMin)
        self.ColorMapMax.colorSelected[QColor].connect(self.getColorMapMax)
        self.bZoom.clicked.connect(self.RubberbandZoom)
        self.bColorMapMin.clicked.connect(self.displayColorMapMin)
        self.bColorMapMax.clicked.connect(self.displayColorMapMax)
        self.bInfo.clicked.connect(self.infoVtkView)
        self.cShowValue.clicked.connect(self.showValues)
        self.setColorSpace()
        self.cViews.currentIndexChanged[int].connect(self.b_loadview)
        self.cViews.lineEdit().editingFinished.connect(self.b_saveview)
        self.cCurrentPath.currentIndexChanged[int].connect(self.changeCurrentPath)
        #      QObject.connect(self.cFamilies,
        #                      SIGNAL("currentIndexChanged(int)"),
        #                      self.changeCurrentFamily)
        self.sIndex1.valueChanged[int].connect(self.highlightPoint)
        self.sIndex2.valueChanged[int].connect(self.highlightPoint)
        self.sIndex3.valueChanged[int].connect(self.highlightPoint)
        self.cShowZone.stateChanged[int].connect(self.fillCurrentPath)
        self.cShowBC.stateChanged[int].connect(self.fillCurrentPath)
        self.cShowCT.stateChanged[int].connect(self.fillCurrentPath)
        self.cShowFamily.stateChanged[int].connect(self.fillCurrentPath)
        self.cRotationAxis.currentIndexChanged[int].connect(self.setInteractor)
        self.cVariables.currentIndexChanged[int].connect(self.getVariable)
        self.cColorSpace.currentIndexChanged[int].connect(self.getColorSpace)
        # QObject.connect(self.sIndex1, SIGNAL("valueChanged(int)"),
        #                 self.highlightPoint)
        # QObject.connect(self.sIndex2, SIGNAL("valueChanged(int)"),
        #                 self.highlightPoint)
        # QObject.connect(self.sIndex3, SIGNAL("valueChanged(int)"),
        #                 self.highlightPoint)
        # QObject.connect(self.cShowZone, SIGNAL("stateChanged(int)"),
        #                 self.fillCurrentPath)
        # QObject.connect(self.cShowBC, SIGNAL("stateChanged(int)"),
        #                 self.fillCurrentPath)
        # QObject.connect(self.cShowCT, SIGNAL("stateChanged(int)"),
        #                 self.fillCurrentPath)
        # QObject.connect(self.cShowFamily, SIGNAL("stateChanged(int)"),
        #                 self.fillCurrentPath)
        # QObject.connect(self.cRotationAxis,
        #                 SIGNAL("currentIndexChanged(int)"),self.setInteractor)
        # QObject.connect(self.cVariables,
        #                 SIGNAL("currentIndexChanged(int)"),
        #                 self.getVariable)
        # QObject.connect(self.cColorSpace,
        #                 SIGNAL("currentIndexChanged(int)"),
        #                 self.getColorSpace)
        self.bResetCamera.clicked.connect(self.ResetCam)

        self.Rotating = 0
        self.Zooming = 0
        self.Panning = 0
        self.Rolling = 0
        self.InteractorIndex = 0
        self.txt = None
        self.observer = None
        self.actorval = None
        self.planeActor = None

    @Slot(str, name='printMessage')
    def printMessage(self, text):
        sys.stdout.write(text + '\n')
        sys.stdout.flush()

    def doRelease(self):
        self._tree = None
        self.reject()

    def parseDone(self):
        print('PARSE DONE')

    def updateSelectedFromVTK(self):
        s = []
        paths = set()
        if self._lastSelectedType == 'Zone':
            for a in self._selected:
                paths.add(CGU.stackPathItem('/', *CGU.getPathToList(a[0], True)[:2]))
        elif self._lastSelectedType == 'Family':
            for a in self._selected:
                paths.add(CGU.stackPathItem('/', self._lastSelectedTag))
        else:
            for a in self._selected:
                paths.add(a[0])
        for p in list(paths):
            if (len(p) > SIZE_PATTERN) and (p[-SIZE_PATTERN:] in LIST_PATTERN):
                p = p[:-SIZE_PATTERN]
            s += ['/' + CGK.CGNSTree_s + p]
        print(s)
        self._tmodel.markExtendToList(s)
        self._tmodel.updateSelected()
        self._parent.treeview.refreshView()

    def setCutPlane(self):
        if self._currentactor is None:
            return
        if not vtk.vtkStructuredGrid().SafeDownCast(self._currentactor[1].GetMapper().GetInput()):
            return
        bounds = self._currentactor[1].GetBounds()
        bds = [0, 0, 0]
        bds[0] = (bounds[0] + bounds[1]) / 2.0
        bds[1] = (bounds[3] + bounds[2]) / 2.0
        bds[2] = (bounds[5] + bounds[4]) / 2.0
        grid = self._currentactor[1].GetMapper().GetInput()
        filter = vtk.vtkStructuredGridGeometryFilter()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        filter.SetInputData(grid)
        self.planeWidget.SetPlaceFactor(1.0)
        self.planeWidget.GetOutlineProperty().SetColor(0, 1, 1)
        self.planeWidget.OutlineTranslationOff()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        self.planeWidget.SetInputConnection(filter.GetOutputPort())
        self.planeWidget.PlaceWidget()
        self.planeWidget.SetOrigin(bds[0], bds[1], bds[2])
        self.planeWidget.SetNormal(1, 0, 0)
        self.planeWidget.On()
        self.planeWidget.AddObserver("InteractionEvent", self.myCallback)
        self._iren.Render()

    def myCallback(self, *args):
        if self.planeWidget.GetDrawPlane() == 0:
            self.planeWidget.DrawPlaneOn()

    def cutting(self):
        if (self._currentactor is None) or (self.planeWidget.GetEnabled() == 0):
            return
        if self.planeActor is not None:
            self._vtkren.RemoveActor(self.planeActor)
        variable = self.cVariables.currentText()
        origin = self.planeWidget.GetOrigin()
        normal = self.planeWidget.GetNormal()
        plane = vtk.vtkPlane()
        plane.SetOrigin(origin)
        plane.SetNormal(normal)
        grid = self._currentactor[1].GetMapper().GetInput()
        filter = vtk.vtkStructuredGridGeometryFilter()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        filter.SetInputData(grid)
        cutter = vtk.vtkCutter()
        cutter.SetCutFunction(plane)
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        filter.SetInputData(grid)
        cutter.Update()
        cutterMapper = vtk.vtkPolyDataMapper()
        cutterMapper.SetInputConnection(cutter.GetOutputPort())
        cutterMapper.SetScalarModeToUseCellData()
        cutterMapper.ScalarVisibilityOn()
        cutterMapper.SetLookupTable(self.lut)
        self.planeActor = vtk.vtkActor()
        self.planeActor.GetProperty().SetColor(1, 1, 0)
        self.planeActor.GetProperty().SetLineWidth(15)
        self.planeActor.GetProperty().SetRepresentationToSurface()
        self.planeActor.SetMapper(cutterMapper)
        self.planeActor.GetMapper().GetInput().GetCellData().SetActiveScalars(variable)
        self.planeWidget.DrawPlaneOff()
        self._vtkren.AddActor(self.planeActor)
        self._iren.Render()

    def infoVtkView(self):
        self._control.helpWindow('VTK')

    def mouseReleaseEvent(self, ev):
        self._ActiveButton = ev.button()
        try:
            ctrl = self._iren.GetControlKey()
        except AttributeError:
            return
        shift = self._iren.GetShiftKey()
        self._iren.SetEventInformationFlipY(ev.x(), ev.y(),
                                            ctrl, shift, chr(0), 0, None)
        if ev.button() == Qt.LeftButton:
            self._iren.LeftButtonReleaseEvent()
            self.Rotating = 0
            self.Panning = 0
            self.Rolling = 0
        elif ev.button() == Qt.RightButton:
            self._iren.RightButtonReleaseEvent()
            self.Zooming = 0
        elif ev.button() == Qt.MidButton:
            self._iren.MiddleButtonReleaseEvent()
            self.Panning = 0

    def setInteractor(self, *args):
        if self.cRotationAxis.currentIndex() == 0:
            self._vtk.SetInteractorStyle(Q7InteractorStyle(self))
        elif self.InteractorIndex == 0:
            self._vtk.SetInteractorStyle(Q7InteractorStyleTrackballObject(self))
        self.InteractorIndex = self.cRotationAxis.currentIndex()

    def setCameraMode(self):
        self.cRotationAxis.clear()
        self.cRotationAxis.addItem("Camera axis all")
        self.cRotationAxis.addItem("Object axis all")
        self.cRotationAxis.addItem("Object axis x")
        self.cRotationAxis.addItem("Object axis y")
        self.cRotationAxis.addItem("Object axis z")

    def ResetCam(self, *args):
        if self.OutlineActor is None:
            self._vtkren.ResetCamera()
        else:
            v = self.OutlineActor.GetBounds()
            print('BOUNDS ', v)
            self._vtkren.ResetCamera(v)
        self._iren.Render()

    def RubberbandZoom(self, *args):
        if self.bZoom.isChecked():
            self._vtk.SetInteractorStyle(Q7InteractorStyleRubberBandZoom(self))
        else:
            interactor = self.cRotationAxis.currentIndex()
            if interactor == 0:
                self._vtk.SetInteractorStyle(Q7InteractorStyle(self))
            else:
                self._vtk.SetInteractorStyle(Q7InteractorStyleTrackballObject(self))

    def showValues(self, *args):
        if self._currentactor is not None:
            if not vtk.vtkStructuredGrid().SafeDownCast(self._currentactor[1].GetMapper().GetInput()):
                self.cShowValue.setChecked(0)
                return
        if (self.cVariables.currentIndex() == 0) or self.OutlineActor is None:
            self.cShowValue.setChecked(0)
            return
        if self.cShowValue.isChecked():
            self.observer = self._iren.AddObserver("MouseMoveEvent", self.displayScalars)
        elif self.observer is not None:
            self._iren.RemoveObserver(self.observer)

    def displayScalars(self, *args):
        self.setPickableOff()
        eventPos = self._iren.GetEventPosition()
        picker = vtk.vtkCellPicker()
        picker.Pick(eventPos[0], eventPos[1], 0.0, self._vtkren)
        cell = picker.GetCellId()
        if cell > -1:
            array = self.cVariables.currentIndex() - 1
            grid = self._currentactor[1].GetMapper().GetInput()
            data = grid.GetCellData().GetArray(array)
            value = data.GetTuple1(cell)
            value = "%.3f" % value
            txt = vtk.vtkTextActor()
            txt.SetInput(value)
            txtprop = txt.GetTextProperty()
            txtprop.SetFontFamilyToArial()
            txtprop.SetFontSize(18)
            txtprop.SetColor(self.color)
            txt.SetDisplayPosition(eventPos[0], eventPos[1])
            self.selectionCellId(grid, cell)
            self._vtkren.AddActor(txt)
            if self.txt is not None:
                self._vtkren.RemoveActor(self.txt)
                self._iren.Render()
            self.txt = txt
        else:
            if self.txt is not None:
                self._vtkren.RemoveActor(self.txt)
            self._vtkren.RemoveActor(self.actorpt)
            self._iren.Render()
        self.setPickableOn()

    def setInteractive(self, *args):
        if self.selectable.isChecked():
            if self.scalarbarwidget is not None:
                self.scalarbarwidget.RepositionableOn()
                self.scalarbarwidget.ResizableOn()
            self.widget.InteractiveOn()

        else:
            if self.scalarbarwidget is not None:
                self.scalarbarwidget.RepositionableOff()
                self.scalarbarwidget.ResizableOff()
            self.widget.InteractiveOff()

    def displayColorMapMin(self, *args):
        self.ColorMapMin.show()

    def displayColorMapMax(self, *args):
        self.ColorMapMax.show()

    def getColorMapMin(self, *args):
        self.maxcolor = self.ColorMapMin.currentColor().getRgbF()
        self.getVariable()

    def getColorMapMax(self, *args):
        self.mincolor = self.ColorMapMax.currentColor().getRgbF()
        self.getVariable()

    def setVariable(self, variable=None):
        self.cVariables.clear()
        self.cVariables.addItem("")
        if variable is not None:
            for i in variable:
                self.cVariables.addItem(i)

    def getVariable(self, *args):
        if self.planeWidget.GetEnabled() == 1:
            self.planeWidget.Off()
        index = self.cVariables.currentIndex()
        if (index == 0) or (not self.cShowValue.isChecked()):
            if self.observer is not None:
                self._iren.RemoveObserver(self.observer)
            self.cShowValue.setChecked(0)
        if (self.mincolor is None) or (self.maxcolor is None):
            return
        variable = self.cVariables.currentText()
        name = variable[0:10]
        self.scalarbar.SetTitle(name)
        if index == 0:
            self.setVisibilityGrids(1)
            if self.scalarbarwidget is not None:
                self.scalarbarwidget.Off()
                self.setScalarVisibilityOff()
                self.scalarbarwidget.GetScalarBarRepresentation().SetPosition(0.9, 0.4)
                self.cShowValue.setChecked(0)
        else:
            a = self._vtkren.GetActors()
            a.InitTraversal()
            s = a.GetNextItem()
            minrange = []
            maxrange = []
            while s:
                grid = s.GetMapper().GetInput()
                tgrid = self.findObjectDims(grid)[0]
                if tgrid == 0:
                    grid.GetCellData().SetActiveScalars(variable)
                    minrange += [grid.GetScalarRange()[0]]
                    maxrange += [grid.GetScalarRange()[1]]
                    s.GetMapper().SetScalarModeToUseCellData()
                    s.GetMapper().ScalarVisibilityOn()
                    if self.lut is not None:
                        s.GetMapper().SetLookupTable(self.lut)
                        self.lut.RemoveAllPoints()
                        self.lut.AddRGBPoint(min(minrange), self.mincolor[0], self.mincolor[1], self.mincolor[2])
                        self.lut.AddRGBPoint(max(maxrange), self.maxcolor[0], self.maxcolor[1], self.maxcolor[2])
                if tgrid == 1:
                    s.VisibilityOff()
                s = a.GetNextItem()
            if self.scalarbarwidget.GetEnabled() == 0:
                sz = self._iren.GetRenderWindow().GetSize()
                self.scalarbarwidget.GetScalarBarRepresentation().SetPosition2(0.07 * 781 / sz[0], 0.6 * 554 / sz[1])
                self.scalarbarwidget.On()
        if self.OutlineActor is not None:
            self.OutlineActor.VisibilityOn()
        self._iren.Render()

    def setVisibilityGrids(self, n):
        a = self._vtkren.GetActors()
        a.InitTraversal()
        s = a.GetNextItem()
        while s:
            grid = s.GetMapper().GetInput()
            if vtk.vtkUnstructuredGrid.SafeDownCast(grid):
                s.SetVisibility(n)
            s = a.GetNextItem()

    def setScalarVisibilityOff(self):
        a = self._vtkren.GetActors()
        a.InitTraversal()
        s = a.GetNextItem()
        while s:
            s.GetMapper().ScalarVisibilityOff()
            s = a.GetNextItem()

    def setColorSpace(self):
        self.cColorSpace.clear()
        self.cColorSpace.addItem("")
        self.cColorSpace.addItem("RGB")
        self.cColorSpace.addItem("HSV")
        self.cColorSpace.addItem("Diverging")

    def getColorSpace(self):
        index = self.cColorSpace.currentIndex()
        if index == 0:
            return
        colorspaces = {1: self.lut.SetColorSpaceToRGB, 2: self.lut.SetColorSpaceToHSV,
                       3: self.lut.SetColorSpaceToDiverging}
        colorspaces[index]()
        self._iren.Render()

    def LookupTable(self):
        self.lut = vtk.vtkColorTransferFunction()
        self.lut.SetColorSpaceToRGB()
        self.lut.SetScaleToLinear()
        text = vtk.vtkTextProperty()
        text.SetFontFamilyToArial()
        text.SetFontSize(12)
        text.SetColor(self.color)
        self.scalarbar = vtk.vtkScalarBarActor()
        self.scalarbar.SetLookupTable(self.lut)
        self.scalarbar.SetNumberOfLabels(5)
        self.scalarbar.SetLabelTextProperty(text)
        self.scalarbar.SetTitleTextProperty(text)
        self.scalarbarwidget = vtk.vtkScalarBarWidget()
        self.scalarbarwidget.ResizableOff()
        self.scalarbarwidget.RepositionableOff()
        self.scalarbarwidget.GetScalarBarRepresentation().SetScalarBarActor(self.scalarbar)
        self.scalarbarwidget.GetScalarBarRepresentation().PickableOff()
        self.scalarbarwidget.GetScalarBarRepresentation().SetPosition(0.9, 0.4)
        self.scalarbarwidget.GetScalarBarRepresentation().SetOrientation(1)

    def screenshot(self):
        sshot = QScreen.grabWindow(self.display.winId())
        sshot.save('/tmp/cg_look-vtkview.png', 'png')

    def findObjectPath(self, selected):
        return self._parser.getPathFromObject(selected)

    def findPathObject(self, path):
        if path in self._cacheActor:
            return self._cacheActor[path]
        alist = self._vtkren.GetActors()
        alist.InitTraversal()
        a = alist.GetNextItem()
        while a:
            if path == self.findObjectPath(a.GetMapper().GetInput()):
                self._cacheActor[path] = a
                return a
            a = alist.GetNextItem()
        return None

    def findObjectDims(self, selected):
        return self._parser.getDimsFromObject(selected)

    #  def leave(self):
    #    self._wtop.destroy()
    #    print self._parent._vtkwindow
    #    self._parent._vtkwindow=None
    #
    #  def onexit(self):
    #    self._control._control.delTreeView(self._viewid,
    #                                       self._fgprint.filedir,
    #                                       self._fgprint.filename)
    #    self.leave()

    def addMarker(self):
        axes = vtk.vtkAxesActor()
        axes.SetShaftTypeToLine()
        axes.SetTotalLength(1, 1, 1)

        self.widget = vtk.vtkOrientationMarkerWidget()
        self.widget.SetOutlineColor(0.93, 0.57, 0.13)
        self.widget.SetOrientationMarker(axes)
        self.widget.SetViewport(0, 0, 0.25, 0.25)

        xLabel = axes.GetXAxisCaptionActor2D()
        xLabel.SetCaption("X")
        xprop = vtk.vtkTextProperty()
        xprop.SetFontSize(3)
        xLabel.SetCaptionTextProperty(xprop)
        xLabel.SetAttachmentPoint(0.75, 0.2, 0)
        xLabel.LeaderOff()
        xLabel.BorderOff()
        xLabel.GetProperty().SetColor(self.color)
        xLabel.SetPosition(0, 0)

        yLabel = axes.GetYAxisCaptionActor2D()
        yLabel.SetCaption("Y")
        yprop = vtk.vtkTextProperty()
        yprop.SetFontSize(3)
        yLabel.SetCaptionTextProperty(yprop)
        yLabel.SetAttachmentPoint(0.2, 0.75, 0)
        yLabel.LeaderOff()
        yLabel.BorderOff()
        yLabel.GetProperty().SetColor(self.color)
        yLabel.SetPosition(0, 0)

        zLabel = axes.GetZAxisCaptionActor2D()
        zLabel.SetCaption("Z")
        zprop = vtk.vtkTextProperty()
        zprop.SetFontSize(3)
        zLabel.SetCaptionTextProperty(zprop)
        zLabel.SetAttachmentPoint(0, 0.2, 0.75)
        zLabel.LeaderOff()
        zLabel.BorderOff()
        zLabel.GetProperty().SetColor(self.color)
        zLabel.SetPosition(0, 0)

        self.xLabel = xLabel
        self.yLabel = yLabel
        self.zLabel = zLabel

    def colourLabel(self):
        self.xLabel.GetProperty().SetColor(self.color)
        self.yLabel.GetProperty().SetColor(self.color)
        self.zLabel.GetProperty().SetColor(self.color)
        self.scalarbar.GetLabelTextProperty().SetColor(self.color)
        self.scalarbar.GetTitleTextProperty().SetColor(self.color)
        self.textwidget.GetTextActor().GetTextProperty().SetColor(self.color)

    def wire(self):
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            actor.GetProperty().SetRepresentationToWireframe()
            actor.GetProperty().EdgeVisibilityOff()
            actor = actors.GetNextItem()

    def addPicker(self):
        txtActor = vtk.vtkActor2D()
        self.txtMapper = vtk.vtkTextMapper()
        txtActor.SetMapper(self.txtMapper)
        self.textActor = vtk.vtkTextActor()
        self.textActor.VisibilityOff()
        self.textActor.PickableOff()
        tprop = vtk.vtkTextProperty()
        tprop.SetFontFamilyToArial()
        tprop.SetFontSize(10)
        tprop.BoldOff()
        tprop.ShadowOff()
        tprop.SetColor(self.color)
        self.textActor.SetTextProperty(tprop)
        self.picker = vtk.vtkCellPicker()
        self.picker.SetTolerance(0.001)
        self.picker.AddObserver("EndPickEvent", self.annotatePick)
        self.textrepresentation = vtk.vtkTextRepresentation()
        self.textrepresentation.GetPositionCoordinate().SetCoordinateSystem(3)
        self.textwidget = vtk.vtkTextWidget()
        self.textwidget.SetRepresentation(self.textrepresentation)
        self.textwidget.SetTextActor(self.textActor)
        self.textwidget.SelectableOff()
        self.textwidget.ResizableOff()

    def annotatePick(self, obj, event):
        if self.wact == 1:
            self.wire()
            self.wact = 0
        if self.controlKey != 1:
            self._selected = []
            self.PropPicked = 0
        if self.picker.GetCellId() < 0:
            self.textActor.VisibilityOff()
            self.textwidget.Off()
            self.highlightProp = 0
        else:
            pickAct = self.picker.GetActors()
            pickAct.InitTraversal()
            a = pickAct.GetNextItem()
            t = ''
            s = None
            while a:
                x = a.GetMapper().GetInput()
                s = self.findObjectPath(x)
                t += s + '\n'
                self._selected += [[s, a]]
                a = pickAct.GetNextItem()
            self.txtMapper.SetInput(t)
            sz = self._iren.GetRenderWindow().GetSize()[1]
            y = sz - self.txtMapper.GetHeight(self._vtkren) - 10.
            h = y / sz
            self.textActor.VisibilityOn()
            self.textActor.SetInput(t)
            self.textActor.SetTextScaleModeToNone()
            self.textActor.GetTextProperty().SetJustificationToLeft()
            self.textrepresentation.SetPosition(0.02, h)
            self.textwidget.On()
            self.PropPicked = 0
            self.controlKey = 0
            self.highlightProp = 1

    def wCGNSTreeParse(self, tree, zlist):
        self._parser = Mesh(tree, zlist)
        self._T('Parse data for VTK display')
        if not self._parser._status:
            MSG.wError(self._control, 500, 'Cannot create VTK view',
                       'No data or bad data for a VTK display')
            return False
        return True

    def wCGNSTreeActors(self, tree, zlist):
        o = vtk.vtkObject()
        o.SetGlobalWarningDisplay(0)
        del o

        self._vtk = self.display
        self._vtk.setParent(self)
        self._vtkren = vtk.vtkRenderer()

        self.addMarker()
        self._vtkwin = self._vtk.GetRenderWindow()
        self._vtkwin.SetNumberOfLayers(1)

        self._vtk.GetRenderWindow().AddRenderer(self._vtkren)

        self._selected = []
        alist = self._parser.createActors()
        for act in alist:
            if act[3] not in self._cacheActor:
                self._cacheActor[act[3]] = act[0]
        variables = []
        if alist != []:
            grid = vtk.vtkStructuredGrid.SafeDownCast(alist[0][2])
            if grid:
                for i in range(grid.GetCellData().GetNumberOfArrays()):
                    variables += [grid.GetCellData().GetArray(i).GetName()]
            for a in alist:
                self._vtkren.AddActor(a[0])
                if a[1] is not None:
                    if self._xmin > a[1][0]:
                        self._xmin = a[1][0]
                    if self._ymin > a[1][2]:
                        self._ymin = a[1][2]
                    if self._zmin > a[1][4]:
                        self._zmin = a[1][4]
                    if self._xmax < a[1][1]:
                        self._xmax = a[1][1]
                    if self._ymax < a[1][3]:
                        self._ymax = a[1][3]
                    if self._zmax < a[1][5]:
                        self._zmax = a[1][5]
        self.setVariable(variables)
        self._vtkren.SetBackground(1, 1, 1)
        self._vtkren.GetActiveCamera().ParallelProjectionOn()
        self._vtkren.ResetCamera()

        (self.vx, self.vy, self.vz) = self._vtkren.GetActiveCamera().GetViewUp()
        (self.cx, self.cy, self.cz) = self._vtkren.GetActiveCamera().GetFocalPoint()
        (self.px, self.py, self.pz) = self._vtkren.GetActiveCamera().GetPosition()
        (self.sx, self.sy, self.sz) = (0., 0., 0.)
        (self.ox, self.oy, self.oz) = (0., 0., 0.)

        self._ctxt = wVTKContext(self._vtkren.GetActiveCamera())
        self._ctxt.setViewUp(self.vx, self.vy, self.vz)
        self._ctxt.setFocalPoint(self.cx, self.cy, self.cz)
        self._ctxt.setPosition(self.px, self.py, self.pz)
        self.LookupTable()
        self.addPicker()
        self._istyle = Q7InteractorStyle(self)
        self._vtk.SetInteractorStyle(self._istyle)
        self._iren = self._istyle.GetInteractor()
        self.scalarbarwidget.SetInteractor(self._iren)
        self.textwidget.SetInteractor(self._iren)
        self.widget.SetInteractor(self._iren)
        self.widget.On()
        self.widget.InteractiveOff()
        self.planeWidget = vtk.vtkImplicitPlaneWidget()
        self.planeWidget.SetInteractor(self._iren)
        self.fillCurrentPath()
        #      self.fillCurrentFamily()

        self._vtkren.AddObserver("StartEvent", self.posWidgets)
        self.setCameraMode()
        self._p_wire = True
        self.setColors(True)

        self._bindings = {'s': self.b_surf,
                          'S': self.b_surf,
                          'q': self.b_surfwire,
                          'Q': self.b_surfwire,
                          'a': self.wireActor,
                          'A': self.wireActor,
                          'w': self.b_wire,
                          'W': self.b_wire,
                          'r': self.resetCam,
                          'R': self.resetCam,
                          'u': self.b_reset,
                          'U': self.b_reset,
                          'd': self.hideActor,
                          'D': self.hideActor}

        return self._vtk.GetRenderWindow()

    def b_shufflecolors(self, pos=None):
        self.setColors(True)

    def b_blackandwhite(self, pos=None):
        self.setColors()

    def getRandomColor(self):
        clst = OCTXT._ColorList
        cl = clst[list(clst)[random.randrange(len(list(clst)))]]
        return cl

    def setColors(self, randcolors=False):
        if not randcolors:
            if self._blackonwhite:
                self._vtkren.SetBackground(1, 1, 1)
                self.color = (0, 0, 0)
                self.colourLabel()
                cl = (0, 0, 0)
            else:
                self._vtkren.SetBackground(0, 0, 0)
                self.color = (1, 1, 1)
                self.colourLabel()
                cl = (1, 1, 1)
            self._blackonwhite = not self._blackonwhite
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            if randcolors:
                cl = self.getRandomColor()
            actor.GetProperty().SetColor(cl)
            actor = actors.GetNextItem()
        if self.OutlineActor is not None:
            self.OutlineActor.GetProperty().SetColor(0, 1, 1)
        if self.highlightProp == 1:
            color = self._currentactor[1].GetProperty().GetColor()
            self._currentactor[1].GetProperty().SetColor(1, 0, 0)
            self._currentactor[2] = color
        else:
            self._currentactor = None
        if self.actorpt is not None:
            self.actorpt.GetProperty().SetColor(0, 1, 0)
        self._iren.Render()

    def b_update_from(self):
        self.busyCursor()
        self._selected = []
        for pth in self._parser.getPathList():
            if pth in self._parent.model().getSelected(True):
                self._selected.append([pth, self.findPathObject(pth)])
        if len(self._selected) > 1:
            MSG.wInfo(self, 503, "VTK view selections",
                      """Only the first marked object in tree view would be
                      selected in the VTK view. That could be one of the
                      Family_t, Zone_t or BC_t SIDS type node.""")
        for s in self._selected:
            self.changeCurrentActor(s)
            break  # todo: allow multiple selections
        self.readyCursor()

    def b_update_to(self):
        self.busyCursor()
        self.textActor.VisibilityOff()
        ptmp = set()
        stmp = []
        for i in self._selected:
            if i[0] not in ptmp:
                ptmp.add(i[0])
                stmp.append(i)
        self._selected = stmp
        self.updateSelectedFromVTK()
        self.fillCurrentPath()
        self.readyCursor()

    def b_reset(self, *args):
        self._selected = []
        for i in self._hidden:
            i[1].VisibilityOn()
        self._hidden = []
        self._selected = []
        self.textActor.VisibilityOff()
        self.changeCurrentActor([None, None])
        self.fillCurrentPath()

    def b_next(self):
        if len(self._selected) > 0:
            self._selected = self._selected[1:] + [self._selected[0]]
            self.PropPicked = 1
            self.changeCurrentActor(self._selected[0], False)
            return self._selected[0]

    def b_prev(self):
        if len(self._selected) > 0:
            self._selected = [self._selected[-1]] + self._selected[0:-1]
            self.PropPicked = 1
            self.changeCurrentActor(self._selected[0], False)

    def b_loadview(self, name=None):
        vname = self.cViews.currentText()
        if vname in self._camera:
            (vu, cr, p, fp, va, ps, pj) = self._camera[vname]
            camera = self._vtkren.GetActiveCamera()
            camera.SetViewUp(vu)
            camera.SetClippingRange(cr)
            camera.SetPosition(p)
            camera.SetFocalPoint(fp)
            camera.SetViewAngle(va)
            camera.SetParallelScale(ps)
            camera.SetParallelProjection(pj)
            self._iren.Render()

    def updateViewList(self):
        k = list(self._camera)
        k.sort()
        self.cViews.clear()
        self.cViews.addItem("")
        for i in k:
            self.cViews.addItem(i)
        self.cViews.setCurrentIndex(0)

    def b_delview(self, name=None):
        name = str(self.cViews.currentText())
        if (name == "") or (name not in self._camera):
            return
        del self._camera[name]
        self.updateViewList()

    def b_saveview(self, name=None):
        camera = self._vtkren.GetActiveCamera()
        name = str(self.cViews.currentText())
        if (name == "") or (name in self._camera):
            return
        self._camera[name] = (camera.GetViewUp(),
                              camera.GetClippingRange(),
                              camera.GetPosition(),
                              camera.GetFocalPoint(),
                              camera.GetViewAngle(),
                              camera.GetParallelScale(),
                              camera.GetParallelProjection()
                              )
        self.updateViewList()

    def b_refresh(self, pos):
        self._vtk.GetRenderWindow().Render()

    def b_saveVTK(self, *args):
        w = vtk.vtkGenericDataObjectWriter()
        w.SetFileName('/tmp/Foo.vtk')
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            if VTK_VERSION_MAJOR < 8:
                raise RuntimeError("VTK version is too old, please upgrade")
            w.SetInputData(actor.GetMapper().GetInput())
            actor = actors.GetNextItem()
        w.Write()

    def b_xaxis(self, pos=None):
        if self.cMirror.isChecked():
            self.setAxis(pos, -1)
        else:
            self.setAxis(pos, 1)

    def b_yaxis(self, pos=None):
        if self.cMirror.isChecked():
            self.setAxis(pos, -2)
        else:
            self.setAxis(pos, 2)

    def b_zaxis(self, pos=None):
        if self.cMirror.isChecked():
            self.setAxis(pos, -3)
        else:
            self.setAxis(pos, 3)

    def changeCurrentFamily(self, *args):
        print(args)

    def fillCurrentFamily(self):
        self.cFamilies.clear()
        #      sel=[n[0] for n in self._fselected]
        #      hid=[n[0] for n in self._fhidden]
        self.cFamilies.addItem(self._epix, '')
        for i in self._parser.getFamilyList():
            pix = self._npix
            #          if (i in sel): pix=self._spix
            #          if (i in hid): pix=self._hpix
            self.cFamilies.addItem(pix, i)
        self.cFamilies.model().sort(0)
        self._iren.Render()

    def setCurrentFamily(self, family):
        pass

    def changeCurrentPath(self, *args):
        path = self.cCurrentPath.currentText()
        if path == '':
            return
        if self.PropPicked == 1:
            self.PropPicked = 0
            return
        actor = self.findPathObject(path)
        self.changeCurrentActor([path, actor])

    def setCurrentPath(self, path):
        ix = self.cCurrentPath.findText(path)
        if ix != -1:
            self.cCurrentPath.setCurrentIndex(ix)

    def fillCurrentPath(self):
        self.cCurrentPath.clear()
        sel = [n[0] for n in self._selected]
        hid = [n[0] for n in self._hidden]
        self.cCurrentPath.addItem(self._epix, '')
        pthlist = []
        if self.cShowZone.isChecked():
            pthlist += ['Zone']
        if self.cShowCT.isChecked():
            pthlist += ['CT']
        if self.cShowBC.isChecked():
            pthlist += ['BC']
        if self.cShowFamily.isChecked():
            pthlist += ['Family']
        for i in self._parser.getPathList(pthlist):
            pix = self._npix
            if i in sel:
                pix = self._spix
            if i in hid:
                pix = self._hpix
            self.cCurrentPath.addItem(pix, i)
        self.cCurrentPath.model().sort(0)
        self._iren.Render()

    def reverseSelection(self):
        selected = []
        selection = []
        hidden = []
        for i in self._selected:
            selected.append(i[0])
        for i in self._hidden:
            hidden.append(i[0])
        for i in self._parser.getPathList():
            if (i not in selected) and (i not in hidden):
                selection.append([i, self.findPathObject(i)])
        self._selected = selection
        actor = self.b_next()
        path = actor[0]
        self.fillCurrentPath()
        if path is not None:
            self.setCurrentPath(path)

    def setAxis(self, pos, iaxis):
        camera = self._vtkren.GetActiveCamera()
        fp = camera.GetFocalPoint()
        pos = camera.GetPosition()
        distance = NPY.sqrt((fp[0] - pos[0]) * (fp[0] - pos[0])
                            + (fp[1] - pos[1]) * (fp[1] - pos[1])
                            + (fp[2] - pos[2]) * (fp[2] - pos[2]))
        if iaxis == 1:
            (vx, vy, vz) = (0., 0., 1.)
            (px, py, pz) = (fp[0] + distance, fp[1], fp[2])
        elif iaxis == 2:
            (vx, vy, vz) = (0., 0., 1.)
            (px, py, pz) = (fp[0], fp[1] + distance, fp[2])
        elif iaxis == 3:
            (vx, vy, vz) = (0., 1., 0.)
            (px, py, pz) = (fp[0], fp[1], fp[2] + distance)
        elif iaxis == -1:
            (vx, vy, vz) = (0., 0., 1.)
            (px, py, pz) = (fp[0] - distance, fp[1], fp[2])
        elif iaxis == -2:
            (vx, vy, vz) = (0., 0., 1.)
            (px, py, pz) = (fp[0], fp[1] - distance, fp[2])
        elif iaxis == -3:
            (vx, vy, vz) = (0., 1., 0.)
            (px, py, pz) = (fp[0], fp[1], fp[2] - distance)
        camera.SetViewUp(vx, vy, vz)
        camera.SetPosition(px, py, pz)
        self._vtkren.ResetCameraClippingRange()
        self._vtkren.Render()
        self._vtkren.ResetCamera()
        self._iren.Render()
        self._ctxt = wVTKContext(camera)
        self._ctxt.setViewUp(vx, vy, vz)
        self._ctxt.setPosition(px, py, pz)

    def b_surf(self, pos):
        if not self._p_wire:
            self.b_wire(pos)
            return
        self._p_wire = False
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            actor.GetProperty().SetRepresentationToSurface()
            actor.GetProperty().EdgeVisibilityOff()
            actor = actors.GetNextItem()
        self._vtk.GetRenderWindow().Render()

    def posWidgets(self, *args):
        if self._iren is not None:
            sz = self._iren.GetRenderWindow().GetSize()
            self.posScalarBar(sz)

    def posScalarBar(self, sz):
        if self.scalarbarwidget.GetEnabled() == 1:
            if self.scalarbarwidget.GetScalarBarActor().GetOrientation() == 1:
                self.scalarbarwidget.GetScalarBarRepresentation().SetPosition2(0.07 * 781 / sz[0], 0.6 * 554 / sz[1])
            else:
                self.scalarbarwidget.GetScalarBarRepresentation().SetPosition2(0.6 * 781 / sz[0], 0.07 * 554 / sz[1])
            pos2 = self.scalarbarwidget.GetScalarBarRepresentation().GetPosition2()
            pos = self.scalarbarwidget.GetScalarBarRepresentation().GetPosition()
            pos = list(pos)
            if pos[0] <= 0:
                pos[0] = 0
            if pos[1] <= 0:
                pos[1] = 0
            if (pos2[1] + pos[1]) >= 1:
                pos[1] = 1 - pos2[1]
                self.scalarbarwidget.GetScalarBarRepresentation().SetPosition(pos[0], 1.2 - pos2[1])
            if (pos2[0] + pos[0]) >= 1.01:
                pos[0] = 1.01 - pos2[0]
            self.scalarbarwidget.GetScalarBarRepresentation().SetPosition(pos[0], pos[1])

    def b_surfwire(self, pos):
        if not self._p_wire:
            self.b_wire(pos)
            return
        self._p_wire = False
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            actor.GetProperty().SetRepresentationToSurface()
            actor.GetProperty().EdgeVisibilityOn()
            actor = actors.GetNextItem()
        if self.OutlineActor is not None:
            self.OutlineActor.GetProperty().EdgeVisibilityOff()
        self._vtk.GetRenderWindow().Render()

    def b_wire(self, pos):
        if self._p_wire:
            self.b_surf(pos)
            return
        self._p_wire = True
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        actor = actors.GetNextItem()
        while actor:
            actor.GetProperty().SetRepresentationToWireframe()
            actor = actors.GetNextItem()
        self._vtk.GetRenderWindow().Render()

    def wireActor(self, pos):
        if not self._selected:
            return
        self.wire()
        for i in self._selected:
            i[1].GetProperty().SetRepresentationToSurface()
            i[1].GetProperty().EdgeVisibilityOff()
        self._vtk.GetRenderWindow().Render()
        self.wact = 1

    def hideActor(self, pos):
        if self._currentactor is not None:
            if self._currentactor[0:2] in self._selected:
                self._currentactor[1].VisibilityOff()
                #              self._currentactor[3].VisibilityOff() ### FIXME
                self._selected.remove(self._currentactor[0:2])
                self._hidden.append(self._currentactor[0:2])
                actor = self.b_next()
                self.fillCurrentPath()
                if actor is not None:
                    s = self.findObjectPath(actor[1].GetMapper().GetInput())
                    if s is not None:
                        self.setCurrentPath(s)
                else:
                    self.changeCurrentActor([None, None])

    def resetCam(self, pos):
        self._vtkren.ResetCamera()
        self._iren.Render()

    def reject(self):
        # some blocking cases may lead to a not-already defined self._master
        # thus closing the window raises an AttributeError
        if self._vtk is not None:
            self._vtk.GetRenderWindow().Finalize()
        self._master._vtkwindow = None

    def changeCurrentActor(self, act, combo=True):
        self._selected = []
        self._lastSelectedType = None
        self._lastSelectedTag = ''
        if act[1] is not None and not act[1].alone():
            self._lastSelectedType = act[1].topo()
            self._lastSelectedTag = act[1].tag()
            atpl = []
            plst = act[1].GetParts()
            plst.InitTraversal()
            a = plst.GetNextItemAsObject()
            while a:
                self._selected += [[a.path(), a]]
                atpl += [(act[0], a)]
                a = plst.GetNextItemAsObject()
        else:
            atpl = [act]
            if act[1] is not None:
                self._lastSelectedType = act[1].topo()
        self.resetSpinBox()
        if self.planeWidget is not None:
            if self.planeWidget.GetEnabled() == 1:
                self.planeWidget.Off()
            if self.planeActor is not None:
                self._vtkren.RemoveActor(self.planeActor)
        if self.actorpt is not None:
            self._vtkren.RemoveActor(self.actorpt)
        if self._currentactor is not None:
            if type(self._currentactor[2]) != list:
                lact = [(self._currentactor[3], self._currentactor[2])]
            else:
                lact = self._currentactor
            if ((lact is not None) and
                    (type(lact) == list) and
                    (lact[0] is not None)):
                for (col, act) in lact[2]:
                    self._vtkren.RemoveActor(act)
                    act.GetProperty().SetColor(col)
                    act.GetProperty().SetLineWidth(1.0)
        aclist = []
        for atp in atpl:
            path = atp[0]
            actor = atp[1]
            if not combo:
                self.setCurrentPath(path)
            if actor is None:
                self.cCurrentPath.setCurrentIndex(0)
                if None not in [self.PickedRenderer, self.OutlineActor]:
                    self.PickedRenderer.RemoveActor(self.OutlineActor)
                    self.PickedRenderer = None
                    self.CurrentRenderer.Render()
                    self.CurrentRenderer.GetRenderWindow().Render()
                return
            self.grid_dims = self.findObjectDims(actor.GetMapper().GetInput())
            color = actor.GetProperty().GetColor()
            actor2 = vtk.vtkActor()
            actor2.ShallowCopy(actor)
            actor2.PickableOff()
            actor2.DragableOff()
            actor2.GetProperty().SetColor(*self.selectionColor)
            actor2.GetProperty().SetLineWidth(3.0)
            self._vtkren.AddActor(actor2)
            self.CurrentRenderer = self._vtkren
            self.Outline = vtk.vtkStructuredGridOutlineFilter()
            self.OutlineMapper = vtk.vtkPolyDataMapper()
            if VTK_VERSION_MAJOR < 8:
                raise RuntimeError("VTK version is too old, please upgrade")
            self.Outline.SetInputData(actor2.GetMapper().GetInput())
            self.OutlineMapper.SetInputConnection(self.Outline.GetOutputPort())
            if None not in [self.PickedRenderer, self.OutlineActor]:
                self.PickedRenderer.RemoveActor(self.OutlineActor)
                self.PickedRenderer = None
            if actor is not None:
                self.OutlineActor = vtk.vtkActor()
                self.OutlineActor.PickableOff()
                self.OutlineActor.DragableOff()
                self.OutlineActor.SetMapper(self.OutlineMapper)
                self.OutlineActor.GetProperty().SetColor(*self.selectionColor)
                self.OutlineActor.GetProperty().SetAmbient(1.0)
                self.OutlineActor.GetProperty().SetDiffuse(0.0)
                self.OutlineActor.GetProperty().SetLineWidth(3.0)
                self.OutlineActor.SetScale(1.01, 1.01, 1.01)
                if self.CurrentRenderer != self.PickedRenderer:
                    if None not in [self.PickedRenderer, self.OutlineActor]:
                        self.PickedRenderer.RemoveActor(self.OutlineActor)
                    self.CurrentRenderer.AddActor(self.OutlineActor)
                    self.PickedRenderer = self.CurrentRenderer
                aclist += [(color, actor2)]
        self._iren.Render()
        self._currentactor = [path, actor, aclist]

    def highlightPoint(self, *args):
        actor = self._currentactor[1]
        if actor is None:
            return
        grid = actor.GetMapper().GetInput()
        tgrid = self.grid_dims[0]
        if tgrid == 0:
            i = self.sIndex1.value()
            j = self.sIndex2.value()
            k = self.sIndex3.value()
            self.s_highlightPoint(grid, (i, j, k))
        if tgrid == 2:
            i = self.sIndex1.value()
            self.uns_highlightPoint(grid, i)
        if tgrid == 1:
            ptid = self.sIndex1.value() - 1
            self.selectionPointId(grid, ptid)

    def s_highlightPoint(self, grid, index):
        if self.actorpt is not None:
            self._vtkren.RemoveActor(self.actorpt)
        if grid is None:
            return
        (i, j, k) = index
        filter = vtk.vtkStructuredGridGeometryFilter()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        filter.SetInputData(grid)
        filter.SetExtent(i - 1, i - 1, j - 1, j - 1, k - 1, k - 1)
        mapper = vtk.vtkPolyDataMapper()
        mapper.SetInputConnection(filter.GetOutputPort())
        self.actorpt = vtk.vtkActor()
        self.actorpt.SetMapper(mapper)
        self.actorpt.GetProperty().SetColor(0, 1, 0)
        self.actorpt.GetProperty().SetPointSize(6)
        self.actorpt.PickableOff()
        self.actorpt.DragableOff()
        self._vtkren.AddActor(self.actorpt)
        self._iren.Render()

    def uns_highlightPoint(self, grid, cellid):
        if self.actorpt is not None:
            self._vtkren.RemoveActor(self.actorpt)
        if grid is None:
            return
        filter = vtk.vtkUnstructuredGridGeometryFilter()
        if VTK_VERSION_MAJOR < 6:
            filter.SetInput(grid)
        else:
            filter.SetInputData(grid)
        filter.CellClippingOn()
        filter.SetCellMinimum(cellid)
        filter.SetCellMaximum(cellid)
        mapper = vtk.vtkDataSetMapper()
        mapper.SetInputConnection(filter.GetOutputPort())
        self.actorpt = vtk.vtkActor()
        self.actorpt.SetMapper(mapper)
        self.actorpt.GetProperty().SetColor(0, 1, 0)
        self.actorpt.GetProperty().SetLineWidth(2)
        self.actorpt.GetProperty().SetRepresentationToWireframe()
        self.actorpt.PickableOff()
        self.actorpt.DragableOff()
        self._vtkren.AddActor(self.actorpt)
        self._iren.Render()

    def selectionCellId(self, grid, ptid):
        if self.actorpt is not None:
            self._vtkren.RemoveActor(self.actorpt)
        if grid is None:
            return
        ids = vtk.vtkIdTypeArray()
        ids.SetNumberOfComponents(1)
        ids.InsertNextValue(ptid)
        selectionNode = vtk.vtkSelectionNode()
        selectionNode.SetFieldType(0)
        selectionNode.SetContentType(4)
        selectionNode.SetSelectionList(ids)
        selection = vtk.vtkSelection()
        selection.AddNode(selectionNode)
        extractSelection = vtk.vtkExtractSelection()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        extractSelection.SetInputData(0, grid)
        extractSelection.SetInputData(1, selection)
        extractSelection.Update()
        selected = vtk.vtkUnstructuredGrid()
        selected.ShallowCopy(extractSelection.GetOutput())
        selectedMapper = vtk.vtkDataSetMapper()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        selectedMapper.SetInputData(selected)
        self.actorpt = vtk.vtkActor()
        self.actorpt.SetMapper(selectedMapper)
        self.actorpt.GetProperty().SetColor(0, 1, 0)
        self.actorpt.GetProperty().SetPointSize(6)
        self.actorpt.PickableOff()
        self.actorpt.DragableOff()
        self._vtkren.AddActor(self.actorpt)
        self._iren.Render()

    def selectionPointId(self, grid, ptid):
        if self.actorpt is not None:
            self._vtkren.RemoveActor(self.actorpt)
        if grid is None:
            return
        ids = vtk.vtkIdTypeArray()
        ids.SetNumberOfComponents(1)
        ids.InsertNextValue(ptid)
        selectionNode = vtk.vtkSelectionNode()
        selectionNode.SetFieldType(1)
        selectionNode.SetContentType(4)
        selectionNode.SetSelectionList(ids)
        selection = vtk.vtkSelection()
        selection.AddNode(selectionNode)
        extractSelection = vtk.vtkExtractSelection()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        extractSelection.SetInputData(0, grid)
        extractSelection.SetInputData(1, selection)
        extractSelection.Update()
        selected = vtk.vtkUnstructuredGrid()
        selected.ShallowCopy(extractSelection.GetOutput())
        selectedMapper = vtk.vtkDataSetMapper()
        if VTK_VERSION_MAJOR < 8:
            raise RuntimeError("VTK version is too old, please upgrade")
        selectedMapper.SetInputData(selected)
        self.actorpt = vtk.vtkActor()
        self.actorpt.SetMapper(selectedMapper)
        self.actorpt.GetProperty().SetColor(0, 1, 0)
        self.actorpt.GetProperty().SetPointSize(6)
        self.actorpt.PickableOff()
        self.actorpt.DragableOff()
        self._vtkren.AddActor(self.actorpt)
        self._iren.Render()

    def setIndexPoint3(self, grid, index):
        (i, j, k) = index
        (idim, jdim, kdim) = self.grid_dims[1]
        self.sIndex1.blockSignals(True)
        self.sIndex2.blockSignals(True)
        self.sIndex3.blockSignals(True)
        self.sIndex1.setRange(1, idim)
        self.sIndex1.setSingleStep(1)
        self.sIndex2.setRange(1, jdim)
        self.sIndex2.setSingleStep(1)
        self.sIndex3.setRange(1, kdim)
        self.sIndex3.setSingleStep(1)
        self.sIndex1.setValue(i)
        self.sIndex2.setValue(j)
        self.sIndex3.setValue(k)
        self.sIndex1.blockSignals(False)
        self.sIndex2.blockSignals(False)
        self.sIndex3.blockSignals(False)
        self.s_highlightPoint(grid, index)

    def setIndexCell1(self, grid, index):
        idmax = grid.GetNumberOfCells()
        self.sIndex1.blockSignals(True)
        self.sIndex1.setRange(1, idmax)
        self.sIndex1.setSingleStep(1)
        self.sIndex1.setValue(index)
        self.sIndex1.blockSignals(False)
        self.uns_highlightPoint(grid, index)

    def setIndexPoint1(self, grid, index):
        idmax = grid.GetNumberOfPoints()
        self.sIndex1.blockSignals(True)
        self.sIndex1.setRange(1, idmax)
        self.sIndex1.setSingleStep(1)
        self.sIndex1.setValue(index + 1)
        self.sIndex1.blockSignals(False)
        self.selectionPointId(grid, index)

    def setPickableOff(self):
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        a = actors.GetNextActor()
        a.PickableOff()
        while a:
            a.PickableOff()
            a = actors.GetNextActor()
        self._currentactor[1].PickableOn()

    def setPickableOn(self):
        actors = self._vtkren.GetActors()
        actors.InitTraversal()
        a = actors.GetNextActor()
        a.PickableOn()
        while a:
            a.PickableOn()
            a = actors.GetNextActor()
        if self._currentactor is not None:
            pass
            # self._currentactor[3].PickableOff() ### FIXME

    def resetSpinBox(self):
        self.sIndex1.blockSignals(True)
        self.sIndex2.blockSignals(True)
        self.sIndex3.blockSignals(True)
        self.sIndex1.setRange(0, 0)
        self.sIndex2.setRange(0, 0)
        self.sIndex3.setRange(0, 0)
        self.sIndex1.setValue(0)
        self.sIndex2.setValue(0)
        self.sIndex3.setValue(0)
        self.sIndex1.blockSignals(False)
        self.sIndex2.blockSignals(False)
        self.sIndex3.blockSignals(False)

    def removeElement(self):
        if self.actorpt is None:
            return
        self._vtkren.RemoveActor(self.actorpt)
        self._iren.Render()
        self.resetSpinBox()

    def keycode(self, *args):
        keycode = self._iren.GetKeyCode()
        control = self._iren.GetControlKey()
        vtkkeys = ['f', 'F', 'r', 'R']
        keys = ['d', 'D', 's', 'S', 'w', 'W', 'q', 'Q', 'a', 'A', 'u', 'U']
        if keycode in vtkkeys:
            self.OnChar()
        if keycode in keys:
            self.CharCallback()
        if keycode in ['z', 'Z', ' ']:
            self.setPick()
        if keycode == 'p' or keycode == 'P':
            self.pickElement()
        if control == 1:
            self.controlKey = 1

    def pickElement(self):
        if (self._selected == []) or self.cShowValue.isChecked():
            return
        grid = self._currentactor[1].GetMapper().GetInput()
        tgrid = self.grid_dims[0]
        eventPos = self._iren.GetEventPosition()
        if tgrid == 0:
            pointid = self.getPointId(eventPos)
            if pointid > -1:
                array = grid.GetPointData().GetArray(0).GetTuple3(pointid)
                self.setIndexPoint3(grid, array)
            else:
                self.removeElement()
        if tgrid == 2:
            cellid = self.getCellId(eventPos)
            if cellid > -1:
                self.setIndexCell1(grid, cellid)
            else:
                self.removeElement()
        if tgrid == 1:
            pointid = self.getPointId(eventPos)
            if pointid > -1:
                self.setIndexPoint1(grid, pointid)
            else:
                self.removeElement()

    def setPick(self):
        self.setPickableOn()
        path = None
        eventPos = self._iren.GetEventPosition()
        self._iren.SetPicker(self.picker)
        if self.picker is not None:
            self.picker.Pick(eventPos[0], eventPos[1], 0.0, self._vtkren)
            pathpick = self.picker.GetPath()
            if pathpick is None:
                self.changeCurrentActor([None, None])
                self.OutlineActor = None
                self.PropPicked = 0
            else:
                actor = pathpick.GetFirstNode().GetViewProp()
                path = self.findObjectPath(actor.GetMapper().GetInput())
                self.changeCurrentActor([path, actor])
                self.PropPicked = 1
            self.fillCurrentPath()
            if path is not None:
                self.setCurrentPath(path)

    def getPointId(self, eventPos):
        picker = vtk.vtkPointPicker()
        picker.SetTolerance(0.005)
        self._iren.SetPicker(picker)
        self.setPickableOff()
        picker.Pick(eventPos[0], eventPos[1], 0.0, self._vtkren)
        pointid = picker.GetPointId()
        return pointid

    def getCellId(self, eventPos):
        picker = vtk.vtkCellPicker()
        picker.SetTolerance(0.001)
        self._iren.SetPicker(picker)
        self.setPickableOff()
        picker.Pick(eventPos[0], eventPos[1], 0.0, self._vtkren)
        cellid = picker.GetCellId()
        return cellid

    def CharCallback(self, *args):
        keycode = self._iren.GetKeyCode()
        control = self._iren.GetControlKey()
        pos = self._iren.GetEventPosition()
        if keycode in self._bindings:
            self._bindings[keycode](pos)
        if control == 1:
            self.controlKey = 1
        return


# -----------------------------------------------------------------------------
class Q7InteractorStyle(vtk.vtkInteractorStyleTrackballCamera):
    def __init__(self, parent):
        self._parent = parent
        self.AddObserver("CharEvent", self.keycode)

    def keycode(self, *args):
        keycode = self._parent._iren.GetKeyCode()
        control = self._parent._iren.GetControlKey()
        vtkkeys = ['f', 'F', 'r', 'R']
        keys = ['d', 'D', 's', 'S', 'w', 'W', 'q', 'Q', 'a', 'A', 'u', 'U']
        if keycode in vtkkeys:
            self.OnChar()
        if keycode in keys:
            self._parent.CharCallback()
        if keycode in ['z', 'Z', ' ']:
            self._parent.setPick()
        if keycode == 'p' or keycode == 'P':
            self._parent.pickElement()
        if keycode == 'o' or keycode == 'O':
            self._parent.setCutPlane()
        if keycode == 't' or keycode == 'T':
            self._parent.cutting()
        if control == 1:
            self._parent.controlKey = 1


# -----------------------------------------------------------------------------
class Q7InteractorStyleTrackballObject(vtk.vtkInteractorStyle):
    def __init__(self, parent):
        self._parent = parent
        self.AddObserver("CharEvent", self.keycode)
        self.AddObserver("LeftButtonPressEvent", self.ButtonEvent)
        self.AddObserver("MouseMoveEvent", self.MouseMove)
        self.AddObserver("RightButtonPressEvent", self.ButtonEvent)
        self.AddObserver("MiddleButtonPressEvent", self.ButtonEvent)
        self.AddObserver("MouseWheelForwardEvent", self.MouseWheelForward)
        self.AddObserver("MouseWheelBackwardEvent", self.MouseWheelBackward)

    def MouseWheelForward(self, *args):
        factor = pow(1.1, 2.0)
        self.Zoom(factor)

    def MouseWheelBackward(self, *args):
        factor = pow(1.1, -2.0)
        self.Zoom(factor)

    def keycode(self, *args):
        keycode = self._parent._iren.GetKeyCode()
        control = self._parent._iren.GetControlKey()
        vtkkeys = ['f', 'F', 'r', 'R']
        keys = ['d', 'D', 's', 'S', 'w', 'W', 'q', 'Q', 'a', 'A', 'u', 'U']
        if keycode in vtkkeys:
            self.OnChar()
        if keycode in keys:
            self._parent.CharCallback()
        if keycode in ['z', 'Z', ' ']:
            self._parent.setPick()
        if keycode == 'p' or keycode == 'P':
            self._parent.pickElement()
        if control == 1:
            self._parent.controlKey = 1

    def ButtonEvent(self, obj, event):
        if event == "LeftButtonPressEvent":
            if self._parent._iren.GetShiftKey():
                self._parent.Panning = 1
            elif self._parent._iren.GetControlKey():
                self._parent.Rolling = 1
            else:
                self._parent.Rotating = 1
        if event == "RightButtonPressEvent":
            self._parent.Zooming = 1
        if event == "MiddleButtonPressEvent":
            self._parent.Panning = 1

    def MouseMove(self, obj, event):
        index = self._parent.cRotationAxis.currentIndex()
        axis = {2: [1, 0, 0], 3: [0, 1, 0], 4: [0, 0, 1]}
        if self._parent.Rotating:
            if index == 1:
                self.CameraModeToObjectAllAxis()
            else:
                self.CameraModeToObjectAxis(axis[index])
        if self._parent.Zooming:
            self.Dolly()
        if self._parent.Panning:
            self.Pan()
        if self._parent.Rolling:
            self.Roll()

    def ComputeDisplayCenter(self):
        if self._parent.OutlineActor is None:
            bounds = (self._parent._xmin, self._parent._xmax, self._parent._ymin,
                      self._parent._ymax, self._parent._zmin, self._parent._zmax)
        else:
            bounds = self._parent._currentactor[1].GetBounds()
        center = [0, 0, 0]
        center[0] = (bounds[0] + bounds[1]) / 2.0
        center[1] = (bounds[2] + bounds[3]) / 2.0
        center[2] = (bounds[4] + bounds[5]) / 2.0
        self._parent._vtkren.SetWorldPoint(center[0], center[1], center[2], 1.0)
        self._parent._vtkren.WorldToDisplay()
        pt = self._parent._vtkren.GetDisplayPoint()
        return pt, center

    def CameraModeToObjectAxis(self, axis):
        camera = self._parent._vtkren.GetActiveCamera()
        transform = vtk.vtkTransform()
        transform.Identity()
        (pt, center) = self.ComputeDisplayCenter()
        transform.Translate(center[0], center[1], center[2])
        dx = self._parent._iren.GetLastEventPosition()[0] - self._parent._iren.GetEventPosition()[0]
        dy = self._parent._iren.GetLastEventPosition()[1] - self._parent._iren.GetEventPosition()[1]
        camera.OrthogonalizeViewUp()
        size = self._parent._vtkren.GetSize()
        transform.RotateWXYZ(360.0 * dx / size[0], axis[0], axis[1], axis[2])
        transform.RotateWXYZ(360.0 * dy / size[1], axis[0], axis[1], axis[2])
        transform.Translate(-center[0], -center[1], -center[2])
        camera.ApplyTransform(transform)
        camera.OrthogonalizeViewUp()
        self._parent._vtkren.ResetCameraClippingRange()
        self._parent._vtkren.GetRenderWindow().Render()

    def CameraModeToObjectAllAxis(self):
        camera = self._parent._vtkren.GetActiveCamera()
        transform = vtk.vtkTransform()
        transform.Identity()
        (pt, center) = self.ComputeDisplayCenter()
        transform.Translate(center[0], center[1], center[2])
        dx = self._parent._iren.GetLastEventPosition()[0] - self._parent._iren.GetEventPosition()[0]
        dy = self._parent._iren.GetLastEventPosition()[1] - self._parent._iren.GetEventPosition()[1]
        camera.OrthogonalizeViewUp()
        viewUp = camera.GetViewUp()
        size = self._parent._vtkren.GetSize()
        transform.RotateWXYZ(360.0 * dx / size[0], viewUp[0], viewUp[1], viewUp[2])
        v2 = [0, 0, 0]
        vtk.vtkMath().Cross(camera.GetDirectionOfProjection(), viewUp, v2)
        transform.RotateWXYZ(-360.0 * dy / size[1], v2[0], v2[1], v2[2])
        transform.Translate(-center[0], -center[1], -center[2])
        camera.ApplyTransform(transform)
        camera.OrthogonalizeViewUp()
        self._parent._vtkren.ResetCameraClippingRange()
        self._parent._vtkren.GetRenderWindow().Render()

    def Dolly(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        center = self._parent._vtkren.GetCenter()
        dy = self._parent._iren.GetEventPosition()[1] - self._parent._iren.GetLastEventPosition()[1]
        dyf = 10 * dy / center[1]
        dollyFactor = pow(1.1, dyf)
        self.Zoom(dollyFactor)

    def Zoom(self, factor):
        camera = self._parent._vtkren.GetActiveCamera()
        camera.SetParallelScale(camera.GetParallelScale() / factor)
        if self._parent._iren.GetLightFollowCamera():
            self._parent._vtkren.UpdateLightsGeometryToFollowCamera()
        self._parent._iren.Render()

    def Pan(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        motionVector = [0, 0, 0]
        fp = camera.GetFocalPoint()

        self._parent._vtkren.SetWorldPoint(fp[0], fp[1], fp[2], 1.0)
        self._parent._vtkren.WorldToDisplay()
        dp = self._parent._vtkren.GetDisplayPoint()
        focalDepth = dp[2]
        self._parent._vtkren.SetDisplayPoint(self._parent._iren.GetEventPosition()[0],
                                             self._parent._iren.GetEventPosition()[1], focalDepth)
        self._parent._vtkren.DisplayToWorld()
        newPickPoint = self._parent._vtkren.GetWorldPoint()
        self._parent._vtkren.SetDisplayPoint(self._parent._iren.GetLastEventPosition()[0],
                                             self._parent._iren.GetLastEventPosition()[1], focalDepth)
        self._parent._vtkren.DisplayToWorld()
        oldPickPoint = self._parent._vtkren.GetWorldPoint()
        motionVector[0] = oldPickPoint[0] - newPickPoint[0]
        motionVector[1] = oldPickPoint[1] - newPickPoint[1]
        motionVector[2] = oldPickPoint[2] - newPickPoint[2]

        viewFocus = list(camera.GetFocalPoint())
        viewPoint = list(camera.GetPosition())
        camera.SetFocalPoint(motionVector[0] + viewFocus[0],
                             motionVector[1] + viewFocus[1],
                             motionVector[2] + viewFocus[2])

        camera.SetPosition(motionVector[0] + viewPoint[0],
                           motionVector[1] + viewPoint[1],
                           motionVector[2] + viewPoint[2])
        if self._parent._iren.GetLightFollowCamera():
            self._parent._vtkren.UpdateLightsGeometryToFollowCamera()
        self._parent._iren.Render()

    def Roll(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        transform = vtk.vtkTransform()
        pos = camera.GetPosition()
        fp = camera.GetFocalPoint()
        axis = [fp[i] - pos[i] for i in range(3)]
        (DisplayCenter, Center) = self.ComputeDisplayCenter()
        x1 = self._parent._iren.GetLastEventPosition()[0] - int(DisplayCenter[0])
        x2 = self._parent._iren.GetEventPosition()[0] - int(DisplayCenter[0])
        y1 = self._parent._iren.GetLastEventPosition()[1] - int(DisplayCenter[1])
        y2 = self._parent._iren.GetEventPosition()[1] - int(DisplayCenter[1])
        zCross = x1 * y2 - y1 * x2
        zCross = float(zCross)
        angle = vtk.vtkMath().DegreesFromRadians(zCross / (math.sqrt(float(x1 * x1 + y1 * y1)) *
                                                           math.sqrt(float(x2 * x2 + y2 * y2))))
        transform.Identity()
        transform.Translate(Center[0], Center[1], Center[2])
        transform.RotateWXYZ(angle, axis[0], axis[1], axis[2])
        transform.Translate(-Center[0], -Center[1], -Center[2])
        camera.ApplyTransform(transform)
        camera.OrthogonalizeViewUp()
        self._parent._vtkren.ResetCameraClippingRange()
        self._parent._iren.Render()

        # -----------------------------------------------------------------------------


class Q7InteractorStyleRubberBandZoom(vtk.vtkInteractorStyleRubberBandZoom):
    def __init__(self, parent):
        self._parent = parent
        self.AddObserver("CharEvent", self.keycode)
        self.AddObserver("MouseMoveEvent", self.MouseMove)
        self.AddObserver("LeftButtonPressEvent", self.ButtonEvent)
        self.AddObserver("RightButtonPressEvent", self.ButtonEvent)
        self.AddObserver("MiddleButtonPressEvent", self.ButtonEvent)
        self.AddObserver("MouseWheelForwardEvent", self.MouseWheelForward)
        self.AddObserver("MouseWheelBackwardEvent", self.MouseWheelBackward)

    def MouseWheelForward(self, *args):
        factor = pow(1.1, 2.0)
        self.Zoom(factor)

    def MouseWheelBackward(self, *args):
        factor = pow(1.1, -2.0)
        self.Zoom(factor)

    def keycode(self, *args):
        keycode = self._parent._iren.GetKeyCode()
        control = self._parent._iren.GetControlKey()
        vtkkeys = ['f', 'F', 'r', 'R']
        keys = ['d', 'D', 's', 'S', 'w', 'W', 'q', 'Q', 'a', 'A', 'u', 'U']
        if keycode in vtkkeys:
            self.OnChar()
        if keycode in keys:
            self._parent.CharCallback()
        if keycode in ['z', 'Z', ' ']:
            self._parent.setPick()
        if keycode == 'p' or keycode == 'P':
            self._parent.pickElement()
        if control == 1:
            self._parent.controlKey = 1

    def ButtonEvent(self, obj, event):
        if event == "LeftButtonPressEvent":
            if self._parent._iren.GetShiftKey():
                self._parent.Panning = 1
            elif self._parent._iren.GetControlKey():
                self._parent.Rolling = 1
            else:
                self.OnLeftButtonDown()
        if event == "RightButtonPressEvent":
            self._parent.Zooming = 1
        if event == "MiddleButtonPressEvent":
            self._parent.Panning = 1

    def MouseMove(self, obj, event):
        index = self._parent.cRotationAxis.currentIndex()
        if self._parent.Zooming:
            self.Dolly()
        if self._parent.Panning:
            self.Pan()
        if self._parent.Rolling:
            self.Roll()
        self.OnMouseMove()

    def ComputeDisplayCenter(self):
        if self._parent.OutlineActor is None:
            bounds = (self._parent._xmin, self._parent._xmax, self._parent._ymin,
                      self._parent._ymax, self._parent._zmin, self._parent._zmax)
        else:
            bounds = self._parent._currentactor[1].GetBounds()
        center = [0, 0, 0]
        center[0] = (bounds[0] + bounds[1]) / 2.0
        center[1] = (bounds[2] + bounds[3]) / 2.0
        center[2] = (bounds[4] + bounds[5]) / 2.0
        self._parent._vtkren.SetWorldPoint(center[0], center[1], center[2], 1.0)
        self._parent._vtkren.WorldToDisplay()
        pt = self._parent._vtkren.GetDisplayPoint()
        return pt, center

    def Dolly(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        center = self._parent._vtkren.GetCenter()
        dy = self._parent._iren.GetEventPosition()[1] - self._parent._iren.GetLastEventPosition()[1]
        dyf = 10 * dy / center[1]
        dollyFactor = pow(1.1, dyf)
        self.Zoom(dollyFactor)

    def Zoom(self, factor):
        camera = self._parent._vtkren.GetActiveCamera()
        camera.SetParallelScale(camera.GetParallelScale() / factor)
        if self._parent._iren.GetLightFollowCamera():
            self._parent._vtkren.UpdateLightsGeometryToFollowCamera()
        self._parent._iren.Render()

    def Pan(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        motionVector = [0, 0, 0]
        fp = camera.GetFocalPoint()

        self._parent._vtkren.SetWorldPoint(fp[0], fp[1], fp[2], 1.0)
        self._parent._vtkren.WorldToDisplay()
        dp = self._parent._vtkren.GetDisplayPoint()
        focalDepth = dp[2]
        self._parent._vtkren.SetDisplayPoint(self._parent._iren.GetEventPosition()[0],
                                             self._parent._iren.GetEventPosition()[1], focalDepth)
        self._parent._vtkren.DisplayToWorld()
        newPickPoint = self._parent._vtkren.GetWorldPoint()
        self._parent._vtkren.SetDisplayPoint(self._parent._iren.GetLastEventPosition()[0],
                                             self._parent._iren.GetLastEventPosition()[1], focalDepth)
        self._parent._vtkren.DisplayToWorld()
        oldPickPoint = self._parent._vtkren.GetWorldPoint()
        motionVector[0] = oldPickPoint[0] - newPickPoint[0]
        motionVector[1] = oldPickPoint[1] - newPickPoint[1]
        motionVector[2] = oldPickPoint[2] - newPickPoint[2]

        viewFocus = list(camera.GetFocalPoint())
        viewPoint = list(camera.GetPosition())
        camera.SetFocalPoint(motionVector[0] + viewFocus[0],
                             motionVector[1] + viewFocus[1],
                             motionVector[2] + viewFocus[2])

        camera.SetPosition(motionVector[0] + viewPoint[0],
                           motionVector[1] + viewPoint[1],
                           motionVector[2] + viewPoint[2])
        if self._parent._iren.GetLightFollowCamera():
            self._parent._vtkren.UpdateLightsGeometryToFollowCamera()
        self._parent._iren.Render()

    def Roll(self, *args):
        camera = self._parent._vtkren.GetActiveCamera()
        transform = vtk.vtkTransform()
        pos = camera.GetPosition()
        fp = camera.GetFocalPoint()
        axis = [fp[i] - pos[i] for i in range(3)]
        (DisplayCenter, Center) = self.ComputeDisplayCenter()
        x1 = self._parent._iren.GetLastEventPosition()[0] - int(DisplayCenter[0])
        x2 = self._parent._iren.GetEventPosition()[0] - int(DisplayCenter[0])
        y1 = self._parent._iren.GetLastEventPosition()[1] - int(DisplayCenter[1])
        y2 = self._parent._iren.GetEventPosition()[1] - int(DisplayCenter[1])
        zCross = x1 * y2 - y1 * x2
        zCross = float(zCross)
        angle = vtk.vtkMath().DegreesFromRadians(zCross / (math.sqrt(float(x1 * x1 + y1 * y1)) *
                                                           math.sqrt(float(x2 * x2 + y2 * y2))))
        transform.Identity()
        transform.Translate(Center[0], Center[1], Center[2])
        transform.RotateWXYZ(angle, axis[0], axis[1], axis[2])
        transform.Translate(-Center[0], -Center[1], -Center[2])
        camera.ApplyTransform(transform)
        camera.OrthogonalizeViewUp()
        self._parent._vtkren.ResetCameraClippingRange()
        self._parent._iren.Render()

        # -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (str, bytes, range, dict)

from CGNS.pyCGNSconfig import HAS_PY2

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QButtonGroup
from qtpy.QtGui import (QColor, QPalette)

from CGNS.NAV.wfingerprint import Q7FingerPrint
from CGNS.NAV.Q7ToolsWindow import Ui_Q7ToolsWindow
from CGNS.NAV.wfingerprint import Q7Window
from CGNS.NAV.wquery import Q7Query, Q7SelectionList
from CGNS.NAV.moption import Q7OptionContext as OCTXT
from CGNS.NAV.diff import diffAB
from CGNS.NAV.merge import mergeAB
from CGNS.NAV.mmergetreeview import (Q7TreeMergeModel, TAG_FRONT, TAG_BACK)
from CGNS.NAV.wtree import Q7Tree
from CGNS.NAV.wdifftreeview import Q7Diff
from CGNS.NAV.wmergetreeview import Q7Merge
from CGNS.NAV.mquery import Q7QueryEntry

import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnskeywords as CGK


# -----------------------------------------------------------------
class Q7Criteria(object):
    def __init__(self):
        self.name = ''
        self.sidstype = ''
        self.value = ''
        self.shape = ''
        self.rgx_name = False
        self.not_name = False
        self.rgx_sids = False
        self.not_sids = False


# -----------------------------------------------------------------
class Q7ToolsView(Q7Window, Ui_Q7ToolsWindow):
    def __init__(self, parent, fgprint, master):
        Q7Window.__init__(self, Q7Window.VIEW_TOOLS, parent, None, None)
        self.bClose.clicked.connect(self.reject)
        self.bInfo.clicked.connect(self.infoToolsView)
        self.bDiff.clicked.connect(self.diffAB)
        self.bMerge.clicked.connect(self.mergeAB)
        self._master = master
        self._fgprint = fgprint
        self._model = fgprint.model
        self._treeview = master.treeview
        self.cGroup.currentIndexChanged.connect(self.fillqueries)
        if HAS_PY2:
            self.cQuery.currentTextChanged[unicode].connect(self.checkquery)
        else:
            self.cQuery.currentTextChanged[str].connect(self.checkquery)
        # QObject.connect(self.cGroup,
        #                SIGNAL("currentIndexChanged(int)"),
        #                self.fillqueries)
        # QObject.connect(self.cQuery,
        #                SIGNAL("valueChanged(int)"),
        #                self.checkquery)
        qlist = Q7Query.queriesNamesList()
        qlist.sort()
        for q in qlist:
            self.cQuery.addItem(q)
        gqlist = Q7Query.queriesGroupsList()
        for gq in gqlist:
            self.cGroup.addItem(gq)
        # ix=-1#self.cQuery.findText()self.querymodel.getCurrentQuery())
        # if (ix!=-1): self.cQuery.setCurrentIndex(ix)
        self.bApply.clicked.connect(self.applyquery)
        pal = self.bApplyBox.palette()
        pal.setColor(QPalette.WindowText, QColor('red'))
        self.bApplyBox.setPalette(pal)
        self.lockable(self.bApply)
        self.bOperateDoc.clicked.connect(self.querydoc)
        self.bRunSearch.clicked.connect(self.runsearch)
        if self._control.query is not None:
            ix = self.cQuery.findText(self._control.query,
                                      flags=Qt.MatchStartsWith)
            if ix != -1:
                self.cQuery.setCurrentIndex(ix)
            self.applyquery()
            self.selectionlist()
        self.sLevel.valueChanged[int].connect(self.updateCriteria)
        if HAS_PY2:
            self.cSIDStype.editTextChanged[unicode].connect(self.updateSIDStypeEntry)
        else:
            self.cSIDStype.editTextChanged[str].connect(self.updateSIDStypeEntry)
        # QObject.connect(self.sLevel,
        #                SIGNAL("valueChanged(int)"),
        #                self.updateCriteria)
        # QObject.connect(self.cSIDStype,
        #                SIGNAL("editTextChanged(QString)"),
        #                self.updateSIDStypeEntry)
        self.criteria = []
        for l in range(3):
            self.criteria.append(Q7Criteria())
            self.criteria[-1].name = ''
            self.criteria[-1].sidstype = ''
            self.criteria[-1].value = ''
            self.criteria[-1].shape = ''
            self.criteria[-1].rgx_name = False
            self.criteria[-1].not_name = False
            self.criteria[-1].rgx_sids = False
            self.criteria[-1].not_sids = False
        self.criteria[0].title = 'Ancestor criteria'
        self.criteria[1].title = 'Current node criteria'
        self.criteria[2].title = 'Children criteria'
        self.previousLevel = 1
        self.cSIDStype.addItems([''] + CGK.cgnstypes)
        self.cDataType.addItems([''] + list(CGK.adftypes))
        self.updateCriteria()

    def updateSIDStypeEntry(self):
        print(self.cSIDStype.currentText())

    def updateCriteria(self):
        l = self.sLevel.value()
        p = self.previousLevel
        self.criteria[p].name = self.eName.text()
        self.criteria[p].sidstype = self.cSIDStype.currentText()
        self.criteria[p].rgx_name = self.cRegexpName.isChecked()
        self.criteria[p].not_name = self.cNotName.isChecked()
        self.criteria[p].rgx_sids = self.cRegexpSIDStype.isChecked()
        self.criteria[p].not_sids = self.cNotSIDStype.isChecked()
        self.eName.setText(self.criteria[l].name)
        st = self.criteria[l].sidstype
        self.cSIDStype.setCurrentIndex(self.cSIDStype.findText(st))
        self.cRegexpName.setChecked(self.criteria[l].rgx_name)
        self.cNotName.setChecked(self.criteria[l].not_name)
        self.cRegexpSIDStype.setChecked(self.criteria[l].rgx_sids)
        self.cNotSIDStype.setChecked(self.criteria[l].not_sids)
        self.previousLevel = l
        self.gCriteria.setTitle(self.criteria[l].title)

    def model(self):
        return self._fgprint.model

    def checkquery(self):
        q = self.cQuery.currentText()
        if (Q7Query.getQuery(q) is not None and
                Q7Query.getQuery(q).hasArgs):
            self.eUserVariable.setEnabled(True)
        else:
            self.eUserVariable.setEnabled(False)

    def fillqueries(self):
        g = self.cGroup.currentText()
        self.cQuery.clear()
        for qn in Q7Query.queriesNamesList():
            if (g == '*') or (Q7Query.getQuery(qn).group == g):
                self.cQuery.addItem(qn)

    def querydoc(self):
        q = self.cQuery.currentText()
        if q in Q7Query.queriesNamesList():
            doc = Q7Query.getQuery(q).doc
            self._control.helpWindowDoc(doc)

    def runsearch(self):
        self.updateCriteria()
        s = "n1=True\nn2=True\nn3=True\n"
        s += "t1=True\nt2=True\nt3=True\n"
        if self.criteria[0].name:
            if self.criteria[0].rgx_name:
                s += "n1=(CGU.stringNameMatches(PARENT,'%s'))\n" % \
                     self.criteria[0].name
            else:
                s += "n1=(PARENT[0]=='%s')\n" % self.criteria[0].name
            if self.criteria[0].not_name:
                s += "n1=not n1\n"
        if self.criteria[1].name:
            if self.criteria[1].rgx_name:
                s += "n2=(CGU.stringMatches(NAME,'%s'))\n" % self.criteria[1].name
            else:
                s += "n2=(NAME=='%s')\n" % self.criteria[1].name
            if self.criteria[1].not_name:
                s += "n2=not n2\n"
        if self.criteria[2].name:
            if self.criteria[1].rgx_name:
                s += "for cn in CGU.childrenNames(NODE):\n  n3=n3 or ('%s'==cn[0])\n" % self.criteria[2].name
            else:
                s += "n3=('%s' in CGU.childrenNames(NODE))\n" % self.criteria[2].name
        s += "rn=n1 and n2 and n3\n"
        if self.criteria[0].sidstype:
            if self.criteria[0].rgx_sids:
                s += "t1=(CGU.stringTypeMatches(PARENT,'%s'))\n" % \
                     self.criteria[0].sidstype
            else:
                s += "t1=(PARENT[3]=='%s')\n" % self.criteria[0].sidstype
            if self.criteria[0].not_sids:
                s += "t1=not t1\n"
        if self.criteria[1].sidstype:
            if self.criteria[1].rgx_sids:
                s += "t2=(CGU.stringTypeMatches(NAME,'%s'))\n" % \
                     self.criteria[1].sidstype
            else:
                s += "t2=(SIDSTYPE=='%s')\n" % self.criteria[1].sidstype
            if self.criteria[1].not_sids:
                s += "t2=not t2\n"
        if self.criteria[2].sidstype:
            if self.criteria[1].rgx_sids:
                s += "t3=False\nfor cn in CHILDREN:\n  t3=t3 or (CGU.stringTypeMatches('%s,cn))\n" % self.criteria[
                    2].sidstype
            else:
                s += "t3=False\nfor cn in CHILDREN:\n  t3=t3 or ('%s'==cn[3])\n" % self.criteria[2].sidstype
        s += "rt=t1 and t2 and t3\n"
        s += "RESULT=rn and rt\n"
        q = Q7QueryEntry('TMP', script=s)
        skp = list(self._fgprint.lazy)
        sl = q.run(self._fgprint.tree, self._fgprint.links, skp, False, '',
                   self._model.getSelected())
        self._model.markExtendToList(sl)
        self._model.updateSelected()
        self._treeview.refreshView()

    def applyquery(self):
        q = self.cQuery.currentText()
        v = self.eUserVariable.text()
        if q in ['', ' ']:
            self.unmarkall()
            return
        qry = Q7Query
        if q in qry.queriesNamesList():
            sl = qry.getQuery(q).run(self._fgprint.tree, self._fgprint.links,
                                     list(self._fgprint.lazy),
                                     False, v, self.model().getSelected())
            self.model().markExtendToList(sl)
            self.model().updateSelected()
            if qry.getQuery(q).requireTreeUpdate():
                self.model().modelReset()
        self._treeview.refreshView()

    def reset(self):
        self.cAncestor.clear()
        self.cVersionA.clear()
        r = self._fgprint.getUniqueTreeViewIdList()
        for i in r:
            self.cAncestor.addItem('%.3d' % i)
            self.cVersionA.addItem('%.3d' % i)
        for i in r:
            self.cTreeA.addItem('%.3d' % i)
            self.cTreeB.addItem('%.3d' % i)
            self.cTreeAncestor.addItem('%.3d' % i)
        self.gForce = QButtonGroup()
        self.gForce.addButton(self.rForceA)
        self.gForce.addButton(self.rForceB)
        self.gForce.addButton(self.rForceNone)
        self.rForceNone.setChecked(True)
        self.gAncestor = QButtonGroup()
        self.gAncestor.addButton(self.rAncestorA)
        self.gAncestor.addButton(self.rAncestorB)
        self.gAncestor.addButton(self.rAncestor)
        self.rAncestor.setChecked(True)
        self.ePrefixA.setText('%sA%s' % (TAG_FRONT, TAG_BACK))
        self.ePrefixB.setText('%sB%s' % (TAG_FRONT, TAG_BACK))

    def infoToolsView(self):
        self._control.helpWindow('Tools')

    def show(self):
        self.reset()
        super(Q7ToolsView, self).show()

    def diffAB(self):
        idx_a = int(self.cAncestor.currentText())
        idx_b = int(self.cVersionA.currentText())
        fpa = self._fgprint.getFingerPrint(idx_a)
        fpb = self._fgprint.getFingerPrint(idx_b)
        diag = {}
        diffAB(fpa.tree, fpb.tree, '', 'A', diag, False)
        dw = Q7Diff(self._control, fpa.index, fpb.index, diag)
        dw.show()

    def mergeAB(self):
        idx_a = int(self.cTreeA.currentText())
        idx_b = int(self.cTreeB.currentText())
        fpa = self._fgprint.getFingerPrint(idx_a)
        fpb = self._fgprint.getFingerPrint(idx_b)
        pfx_a = self.ePrefixA.text()
        pfx_b = self.ePrefixB.text()
        tree = CGL.newCGNSTree()
        tc = fpa.control.newtreecount
        fpc = Q7FingerPrint(fpa.control, '.', 'new#%.3d.hdf' % tc, tree, [], [])
        Q7TreeMergeModel(fpc)
        self.merge = Q7Tree(fpa.control, '/', fpc)
        fpc._status = [Q7FingerPrint.STATUS_MODIFIED]
        fpa.control.newtreecount += 1
        diag = {}
        diffAB(fpa.tree, fpb.tree, '', 'A', diag, False)
        fpc.tree = mergeAB(fpa.tree, fpb.tree, fpc.tree, 'C', diag, pfx_a, pfx_b)
        fpc.model.modelReset()
        dw = Q7Merge(self._control, fpc, diag)
        dw.show()
        self.merge.hide()

    def reject(self):
        if self.merge is not None:
            self.merge.show()
            self.merge = None
        if self._master._control._toolswindow is not None:
            self._master._control._toolswindow = None

    def closeEvent(self, event):
        self.reject()
        event.accept()

# -----------------------------------------------------------------



# -*- coding: utf-8 -*-
#
#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

import shutil
import os
import os.path as op
import sys
import imp
import tempfile

from time import gmtime, strftime
from CGNS.pyCGNSconfig import version as __vid__
from CGNS.pyCGNSconfig import HAS_MSW 

WINOPTS = HAS_MSW

import CGNS.PAT.cgnskeywords as CGK

from qtpy.QtGui import QFont, QFontDatabase, QTextCursor

# User file paths:
conf_path = op.join(op.expanduser('~'), ".CGNS.NAV")
tmp_path = "/tmp"
if sys.platform == 'win32':
    tmp_path = os.environ['TEMP']


# -----------------------------------------------------------------
def removeSubDirAndFiles(path):
    shutil.rmtree(path)


def copyOneFile(src, dst):
    f1 = open(src, 'r')
    l1 = f1.readlines()
    f1.close()
    if op.exists(dst):
        os.remove(dst)
    f2 = open(dst, 'w')
    f2.writelines(l1)
    f2.close()


# -----------------------------------------------------------------
class Q7OptionContext(object):
    Q_VAR_NODE = 'NODE'
    Q_VAR_PARENT = 'PARENT'
    Q_VAR_NAME = 'NAME'
    Q_VAR_VALUE = 'VALUE'
    Q_VAR_CGNSTYPE = 'SIDSTYPE'
    Q_VAR_CHILDREN = 'CHILDREN'
    Q_VAR_TREE = 'TREE'
    Q_VAR_PATH = 'PATH'
    Q_VAR_LINKS = 'LINKS'
    Q_VAR_SKIPS = 'SKIPS'
    Q_VAR_RESULT = 'RESULT'
    Q_VAR_USER = 'ARGS'
    Q_VAR_SELECTED = 'SELECTED'
    Q_VAR_RESULT_LIST = '__Q7_QUERY_RESULT__'

    Q_SCRIPT_PRE = """
import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnslib as CGL
import numpy
"""

    Q_FILE_PRE = """
import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnslib as CGL
import CGNS.NAV.moption as CGO
import numpy
"""

    Q_SCRIPT_POST = """
try:
  %s[0]=%s
except NameError:
  %s[0]=None
""" % (Q_VAR_RESULT_LIST, Q_VAR_RESULT, Q_VAR_RESULT_LIST)

    _depends = {
        'MaxDisplayDataSize': ['DoNotDisplayLargeData'],
        'MaxLoadDataSize': ['DoNotLoadLargeArrays'],
        'MaxRecursionLevel': ['RecursiveTreeDisplay'],
    }
    _HasProPackage = True
    CHLoneTrace = False
    QueryNoException = False
    ActivateMultiThreading = False
    NAVTrace = False
    AutoExpand = False
    RecursiveTreeDisplay = False
    OneViewPerTreeNode = False
    ShowTableIndex = True
    RecursiveSIDSPatternsLoad = True
    DoNotDisplayLargeData = False
    CheckOnTheFly = False
    FollowLinksAtLoad = True
    DoNotFollowLinksAtSave = True
    AddCurrentDirInSearch = True
    AddRootDirInSearch = True
    DoNotLoadLargeArrays = True
    ShowSIDSStatusColumn = True
    ForceSIDSLegacyMapping = False
    ForceFortranFlag = True
    FilterCGNSFiles = True
    FilterHDFFiles = True
    FilterOwnFiles = True
    FileUpdateRemovesChildren = True
    TransposeArrayForView = True
    Show1DAsPlain = True
    ShowSIDSColumn = True
    ShowLinkColumn = True
    ShowSelectColumn = True
    ShowUserColumn = True
    ShowCheckColumn = True
    ShowShapeColumn = True
    ShowDataTypeColumn = True
    UserCSS = ""
    SelectionListDirectory = op.normpath(op.join(conf_path, 'selections'))
    QueriesDirectory = op.normpath(op.join(conf_path, 'queries'))
    FunctionsDirectory = op.normpath(op.join(conf_path, 'functions'))
    SnapShotDirectory = op.normpath(op.join(conf_path, 'snapshots'))
    _HistoryFileName = op.normpath(op.join(conf_path, 'historyfile.py'))
    _OptionsFileName = op.normpath(op.join(conf_path, 'optionsfile.py'))
    _QueriesDefaultFile = 'default.py'
    _FunctionsDefaultFile = 'default.py'
    IgnoredMessages = []
    LinkSearchPathList = []
    ProfileSearchPathList = []
    GrammarSearchPathList = []
    ValKeyList = ['sample']
    CGNSFileExtension = ['.cgns', '.adf']
    HDFFileExtension = ['.hdf', '.hdf5']
    OwnFileExtension = ['.cgh']
    MaxLoadDataSize = 1000
    MaxDisplayDataSize = 1000
    MaxRecursionLevel = 7
    ADFConversionCom = 'cgnsconvert'
    TemporaryDirectory = op.normpath(tmp_path)
    _ConvertADFFiles = True
    _ToolName = 'CGNS.NAV'
    _ToolVersion = '%s' % __vid__

    _CopyrightNotice = u"""
Copyright (c) Marc Poinot <br>
Copyright (c) Safran - www.safran-group.com<br><br>
Copyright (c) Onera - www.onera.fr<br><br>
<b>all other copyrights and used versions listed below</b>

<hr>
<h3>Contributors (alphabetic order)</h3>
<table>
<tr><td>Florent Cayré</td><td>-SNECMA, France</td></tr>
<tr><td>Alexandre Fayolle</td><td>-LOGILAB, France</td></tr>
<tr><td>Xavier Garnaud</td><td>-SAFRAN, France</td></tr>
<tr><td>Loic Hauchard</td><td>-ONERA (Student INSA Rouen, France)</td></tr>
<tr><td>Elise Hénaux</td><td>-ONERA (Student FIIFO Orsay, France)</td></tr>
<tr><td>Grégory Laheurte</td><td>-ONERA, France</td></tr>
<tr><td>Pierre-Jacques Legay</td><td>-BERTIN, France</td></tr>
<tr><td>Bill Perkins</td><td>-Pacific Northwest Ntl Lab, U.S.A.</td></tr>
<tr><td>Mickael Philit</td><td>-SAFRAN, France</td></tr>
<tr><td>Jérôme Regis</td><td>-STILOG, France</td></tr>
<tr><td>Benoit Rodriguez</td><td>-ONERA, France</td></tr>
<tr><td>Tristan Soubrié</td><td>-ANDHEO, France</td></tr>
<tr><td>Francois Thirifays</td><td>-CENAERO, Belgique</td></tr>
<tr><td>Simon Verley</td><td>-ONERA, France</td></tr>
<tr><td>Ching-Yang Wang</td><td>-U.S.A.</td></tr>
</table>

<h2>Copyrights & Versions</h2>
<hr>
%(pycgnsversion)s<br>
All <b>pyCGNS</b> 
rights reserved in accordance with GPL v2 <br><br>
<b>NO WARRANTY :</b><br>
Check GPL v2 sections 15 and 16
about <font color=red>loss of data or corrupted data</font><br>
<hr>
%(pyqtversion)s<br>
PyQt Copyright (c) Riverbank Computing Limited. <br>
Usage within the terms of the GPL v2.<br>
All Rights Reserved.<br>
<hr>
%(qtversion)s<br>
Qt Copyright (c)<br>
The Qt4 Library is (c) 2011 Nokia Corporation and/or its subsidiary(-ies),
and is licensed under the GNU Lesser General Public License version 2.1
with Nokia Qt LGPL exception version 1.1. <br>
<hr>
%(pythonversion)s<br>
Python Copyright (c)<br>
Copyright (c) 2001-2011 Python Software Foundation.<br>
All Rights Reserved.<br>
<br>
Copyright (c) 2000 BeOpen.com.<br>
All Rights Reserved.<br>
<br>
Copyright (c) 1995-2001 Corporation for National Research Initiatives.<br>
All Rights Reserved.<br>
<br>
Copyright (c) 1991-1995 Stichting Mathematisch Centrum, Amsterdam.<br>
All Rights Reserved.<br>
<br>
<hr>
%(numpyversion)s<br>
Numpy Copyright (c)<br>
Copyright (c) 2005, NumPy Developers<br>
<hr>
%(hdf5version)s<br>
HDF5 Copyright (c)<br>
HDF5 (Hierarchical Data Format 5) Software Library and Utilities<br>
Copyright 2006-2013 by The HDF Group.<br>
NCSA HDF5 (Hierarchical Data Format 5) Software Library and Utilities<br>
Copyright 1998-2006 by the Board of Trustees of the University of Illinois.<br>
<hr>
%(vtkversion)s<br>
VTK Copyright (c)<br>
Copyright (c) 1993-2008 Ken Martin, Will Schroeder, Bill Lorensen <br>
All rights reserved.<br>
<hr>
%(cythonversion)s<br>
Cython Copyright (c)<br>
(c) Copyright 2012, Stefan Behnel, Robert Bradshaw, Dag Sverre Seljebotn,
Greg Ewing, William Stein, Gabriel Gellner, et al.
<hr>
Icons Copyright (c)<br>
All these nice icons are provided or are modified versions of original
icons provided by Mark James (Birmingham, UK).<br>
Please visit his web site: http://www.famfamfam.com/<br>

"""
    #__fv = QFont(QFontDatabase().families()[0]) # Spyder-like but needs a QApplication
    __fv = QFont('Courier new')
    __fc = QFont('Courier new')

    _Label_Font = QFont(__fv)
    _Button_Font = QFont(__fv)
    _Table_Font = QFont(__fc)
    _Edit_Font = QFont(__fc)
    _RName_Font = QFont(__fc)
    _NName_Font = QFont(__fc)
    _NName_Font.setBold(True)

    _Default_Fonts = {
        'Label_Family': _Label_Font.family(),
        'Label_Size': _Label_Font.pointSize(),
        'Label_Bold': False,
        'Label_Italic': False,
        'Table_Family': _Table_Font.family(),
        'Table_Size': _Table_Font.pointSize(),
        'Table_Bold': False,
        'Table_Italic': False,
        'Edit_Family': _Edit_Font.family(),
        'Edit_Size': _Edit_Font.pointSize(),
        'Edit_Bold': False,
        'Edit_Italic': False,
        'Button_Family': _Button_Font.family(),
        'Button_Size': _Button_Font.pointSize(),
        'Button_Bold': False,
        'Button_Italic': False,
        'RName_Family': _RName_Font.family(),
        'RName_Size': _RName_Font.pointSize(),
        'RName_Bold': False,
        'RName_Italic': False,
        'NName_Family': _NName_Font.family(),
        'NName_Size': _NName_Font.pointSize(),
        'NName_Bold': False,
        'NName_Italic': True,
    }

    UserColors = [
        'gray',
        'red',
        'green',
        'blue',
        'orange',
        None,
        None,
        None,
        None,
        None,
    ]

    _ColorList = {
        'cold_grey': (0.5000, 0.5400, 0.5300),
        'dim_grey': (0.4118, 0.4118, 0.4118),
        'grey': (0.7529, 0.7529, 0.7529),
        'light_grey': (0.8275, 0.8275, 0.8275),
        'slate_grey': (0.4392, 0.5020, 0.5647),
        'slate_grey_dark': (0.1843, 0.3098, 0.3098),
        'slate_grey_light': (0.4667, 0.5333, 0.6000),
        'warm_grey': (0.5000, 0.5000, 0.4100),
        'black': (0.0000, 0.0000, 0.0000),
        'ivory_black': (0.1600, 0.1400, 0.1300),
        'lamp_black': (0.1800, 0.2800, 0.2300),
        'alizarin_crimson': (0.8900, 0.1500, 0.2100),
        'brick': (0.6100, 0.4000, 0.1200),
        'coral': (1.0000, 0.4980, 0.3137),
        'coral_light': (0.9412, 0.5020, 0.5020),
        'deep_pink': (1.0000, 0.0784, 0.5765),
        'firebrick': (0.6980, 0.1333, 0.1333),
        'geranium_lake': (0.8900, 0.0700, 0.1900),
        'hot_pink': (1.0000, 0.4118, 0.7059),
        'light_salmon': (1.0000, 0.6275, 0.4784),
        'madder_lake_deep': (0.8900, 0.1800, 0.1900),
        'maroon': (0.6902, 0.1882, 0.3765),
        'pink': (1.0000, 0.7529, 0.7961),
        'pink_light': (1.0000, 0.7137, 0.7569),
        'raspberry': (0.5300, 0.1500, 0.3400),
        'rose_madder': (0.8900, 0.2100, 0.2200),
        'salmon': (0.9804, 0.5020, 0.4471),
        'tomato': (1.0000, 0.3882, 0.2784),
        'beige': (0.6400, 0.5800, 0.5000),
        'brown': (0.5000, 0.1647, 0.1647),
        'brown_madder': (0.8600, 0.1600, 0.1600),
        'brown_ochre': (0.5300, 0.2600, 0.1200),
        'burlywood': (0.8706, 0.7216, 0.5294),
        'burnt_sienna': (0.5400, 0.2100, 0.0600),
        'burnt_umber': (0.5400, 0.2000, 0.1400),
        'chocolate': (0.8235, 0.4118, 0.1176),
        'deep_ochre': (0.4500, 0.2400, 0.1000),
        'flesh': (1.0000, 0.4900, 0.2500),
        'flesh_ochre': (1.0000, 0.3400, 0.1300),
        'gold_ochre': (0.7800, 0.4700, 0.1500),
        'greenish_umber': (1.0000, 0.2400, 0.0500),
        'khaki': (0.9412, 0.9020, 0.5490),
        'khaki_dark': (0.7412, 0.7176, 0.4196),
        'light_beige': (0.9608, 0.9608, 0.8627),
        'peru': (0.8039, 0.5216, 0.2471),
        'rosy_brown': (0.7373, 0.5608, 0.5608),
        'raw_sienna': (0.7800, 0.3800, 0.0800),
        'raw_umber': (0.4500, 0.2900, 0.0700),
        'sepia': (0.3700, 0.1500, 0.0700),
        'sienna': (0.6275, 0.3216, 0.1765),
        'saddle_brown': (0.5451, 0.2706, 0.0745),
        'sandy_brown': (0.9569, 0.6431, 0.3765),
        'tan': (0.8235, 0.7059, 0.5490),
        'van_dyke_brown': (0.3700, 0.1500, 0.0200),
        'cadmium_orange': (1.0000, 0.3800, 0.0100),
        'carrot': (0.9300, 0.5700, 0.1300),
        'dark_orange': (1.0000, 0.5490, 0.0000),
        'mars_orange': (0.5900, 0.2700, 0.0800),
        'mars_yellow': (0.8900, 0.4400, 0.1000),
        'orange': (1.0000, 0.5000, 0.0000),
        'orange_red': (1.0000, 0.2706, 0.0000),
        'yellow_ochre': (0.8900, 0.5100, 0.0900),
        'aureoline_yellow': (1.0000, 0.6600, 0.1400),
        'banana': (0.8900, 0.8100, 0.3400),
        'cadmium_lemon': (1.0000, 0.8900, 0.0100),
        'cadmium_yellow': (1.0000, 0.6000, 0.0700),
        'cadmium_yellow_light': (1.0000, 0.6900, 0.0600),
        'gold': (1.0000, 0.8431, 0.0000),
        'goldenrod': (0.8549, 0.6471, 0.1255),
        'goldenrod_dark': (0.7216, 0.5255, 0.0431),
        'goldenrod_light': (0.9804, 0.9804, 0.8235),
        'goldenrod_pale': (0.9333, 0.9098, 0.6667),
        'light_goldenrod': (0.9333, 0.8667, 0.5098),
        'melon': (0.8900, 0.6600, 0.4100),
        'naples_yellow_deep': (1.0000, 0.6600, 0.0700),
        'yellow': (1.0000, 1.0000, 0.0000),
        'yellow_light': (1.0000, 1.0000, 0.8784),
        'chartreuse': (0.4980, 1.0000, 0.0000),
        'chrome_oxide_green': (0.4000, 0.5000, 0.0800),
        'cinnabar_green': (0.3800, 0.7000, 0.1600),
        'cobalt_green': (0.2400, 0.5700, 0.2500),
        'emerald_green': (0.0000, 0.7900, 0.3400),
        'forest_green': (0.1333, 0.5451, 0.1333),
        'green': (0.0000, 1.0000, 0.0000),
        'green_dark': (0.0000, 0.3922, 0.0000),
        'green_pale': (0.5961, 0.9843, 0.5961),
        'green_yellow': (0.6784, 1.0000, 0.1843),
        'lawn_green': (0.4863, 0.9882, 0.0000),
        'lime_green': (0.1961, 0.8039, 0.1961),
        'mint': (0.7400, 0.9900, 0.7900),
        'olive': (0.2300, 0.3700, 0.1700),
        'olive_drab': (0.4196, 0.5569, 0.1373),
        'olive_green_dark': (0.3333, 0.4196, 0.1843),
        'permanent_green': (0.0400, 0.7900, 0.1700),
        'sap_green': (0.1900, 0.5000, 0.0800),
        'sea_green': (0.1804, 0.5451, 0.3412),
        'sea_green_dark': (0.5608, 0.7373, 0.5608),
        'sea_green_medium': (0.2353, 0.7020, 0.4431),
        'sea_green_light': (0.1255, 0.6980, 0.6667),
        'spring_green': (0.0000, 1.0000, 0.4980),
        'spring_green_medium': (0.0000, 0.9804, 0.6039),
        'terre_verte': (0.2200, 0.3700, 0.0600),
        'viridian_light': (0.4300, 1.0000, 0.4400),
        'yellow_green': (0.6039, 0.8039, 0.1961),
        'aquamarine': (0.4980, 1.0000, 0.8314),
        'aquamarine_medium': (0.4000, 0.8039, 0.6667),
        'cyan': (0.0000, 1.0000, 1.0000),
        'cyan_white': (0.8784, 1.0000, 1.0000),
        'turquoise': (0.2510, 0.8784, 0.8157),
        'turquoise_dark': (0.0000, 0.8078, 0.8196),
        'turquoise_medium': (0.2824, 0.8196, 0.8000),
        'turquoise_pale': (0.6863, 0.9333, 0.9333),
        'alice_blue': (0.9412, 0.9725, 1.0000),
        'blue': (0.0000, 0.0000, 1.0000),
        'blue_light': (0.6784, 0.8471, 0.9020),
        'blue_medium': (0.0000, 0.0000, 0.8039),
        'cadet': (0.3725, 0.6196, 0.6275),
        'cobalt': (0.2400, 0.3500, 0.6700),
        'cornflower': (0.3922, 0.5843, 0.9294),
        'cerulean': (0.0200, 0.7200, 0.8000),
        'dodger_blue': (0.1176, 0.5647, 1.0000),
        'indigo': (0.0300, 0.1800, 0.3300),
        'manganese_blue': (0.0100, 0.6600, 0.6200),
        'midnight_blue': (0.0980, 0.0980, 0.4392),
        'navy': (0.0000, 0.0000, 0.5020),
        'peacock': (0.2000, 0.6300, 0.7900),
        'powder_blue': (0.6902, 0.8784, 0.9020),
        'royal_blue': (0.2549, 0.4118, 0.8824),
        'slate_blue': (0.4157, 0.3529, 0.8039),
        'slate_blue_dark': (0.2824, 0.2392, 0.5451),
        'slate_blue_light': (0.5176, 0.4392, 1.0000),
        'slate_blue_medium': (0.4824, 0.4078, 0.9333),
        'sky_blue': (0.5294, 0.8078, 0.9216),
        'sky_blue_deep': (0.0000, 0.7490, 1.0000),
        'sky_blue_light': (0.5294, 0.8078, 0.9804),
        'steel_blue': (0.2745, 0.5098, 0.7059),
        'steel_blue_light': (0.6902, 0.7686, 0.8706),
        'turquoise_blue': (0.0000, 0.7800, 0.5500),
        'ultramarine': (0.0700, 0.0400, 0.5600),
        'blue_violet': (0.5412, 0.1686, 0.8863),
        'cobalt_violet_deep': (0.5700, 0.1300, 0.6200),
        'magenta': (1.0000, 0.0000, 1.0000),
        'orchid': (0.8549, 0.4392, 0.8392),
        'orchid_dark': (0.6000, 0.1961, 0.8000),
        'orchid_medium': (0.7294, 0.3333, 0.8275),
        'permanent_red_violet': (0.8600, 0.1500, 0.2700),
        'plum': (0.8667, 0.6275, 0.8667),
        'purple': (0.6275, 0.1255, 0.9412),
        'purple_medium': (0.5765, 0.4392, 0.8588),
        'ultramarine_violet': (0.3600, 0.1400, 0.4300),
        'violet': (0.5600, 0.3700, 0.6000),
        'violet_dark': (0.5804, 0.0000, 0.8275),
        'violet_red': (0.8157, 0.1255, 0.5647),
        'violet_red_medium': (0.7804, 0.0824, 0.5216),
        'violet_red_pale': (0.8588, 0.4392, 0.5765),
    }

    _ReservedNames = CGK.cgnsnames
    _ReservedTypes = CGK.cgnstypes
    _SortedTypeList = CGK.sortedtypelist

    _UsualQueries = [

# ---------------------------------------------------------------------------
# INDENTATION IS SIGNIFICANT
# ---------------------------------------------------------------------------
    
# --- Search -----------------------------------------------------------
# last two booleans: Update tree, has args
('001. Node name',
'Search by',
'RESULT=(NAME==ARGS[0])',
"""
Search by
Node name

Search all nodes with the exact NAME as argument.

The argument name need not to be a tuple or to have quotes,
all the following values are ok and would match the NAME <i>ZoneType</i>:

ZoneType
'ZoneType'
('ZoneType',)
""", False, True),

('002. Wildcard node name',
'Search by',
"""import fnmatch
RESULT=fnmatch.fnmatchcase(NAME,ARGS[0])
""",
"""
Search by
Wildcard node name

Search all nodes with the wildcard NAME as argument.

Warning: the <b>argument name</b> should be quoted:

'BC*' is ok

BC* would fail
""", False, True),

('003. Node type',
'Search by',
'RESULT=(SIDSTYPE==ARGS[0])',
"""search all nodes with argument SIDS type.""", False, True),

('005. Node value',
'Search by',
"""
from numpy import *
target=eval(ARGS[0])
if   (VALUE is None and target is None): RESULT=True
elif (VALUE is None)            : RESULT=False
elif (target.dtype!=VALUE.dtype): RESULT=False
elif (target.size!=VALUE.size):   RESULT=False
elif (target.shape!=VALUE.shape): RESULT=False
elif (target.tolist()==VALUE.tolist()): RESULT=True
else:                             RESULT=False
""",
"""search all nodes with argument value. The compare is performed
using a straightforward '==' and then relies on the Python/Numpy comparison
operator.""", False, True),

('010. Node with truncated data',
'Search by',
'if (PATH in SKIPS): RESULT=PATH',
"""search all nodes with truncated or unread data, for example if you have set
the maximum data argument for the load, or if you release the memory of a
node.""", False, False),

('004. Wildcard node type',
'Search by',
"""
import fnmatch
RESULT=fnmatch.fnmatchcase(SIDSTYPE,ARGS[0])
""",
"""
Search by
Wildcard node type

Search all nodes with wildcard argument SIDS type.
Warning: the <b>argument type</b> should be quoted:

'Turb*' is ok

Turb* would fail
""", False, True),

('011. Non-MT UserDefinedData',
'Search by',
"""
if (    (SIDSTYPE==CGK.UserDefinedData_ts)
    and (CGU.getValueDataType(NODE)!=CGK.MT)):
  RESULT=True
""",
"""
Search by
Valued UserDefinedData

Search all <b>UserDefinedData_t</b> nodes with a non-<b>MT</b> data type.

No argument.
""", False, False),
# -----------------------------------------------------------------------------
('012. FamilyName',
'Search by',
"""
if (SIDSTYPE in [CGK.FamilyName_ts, CGK.AdditionalFamilyName_ts]):
    RESULT=True
""",
"""
Search by
All <b>FamilyName_t</b> and <b>AdditionalFamilyname_t</b> nodes.
""", False, False),
# -----------------------------------------------------------------------------
('013. FamilyName reference',
'Search by',
"""
if ((SIDSTYPE in [CGK.FamilyName_ts, CGK.AdditionalFamilyName_ts]) and
    (VALUE.tostring().decode('ascii')==ARGS[0])):
    RESULT=True
""",
"""
Search by<br>
Reference to a FamilyName<br>
 
Search all <b>FamilyName</b> nodes with the arg string (plain).<br>
The string arg should be a valid Python string such as:<br>
 
'BLADE(1)'<br>
'Surface ext 1A'<br>
'Left/Wing/Flap'<br>
 
""", False, True),
# -----------------------------------------------------------------------------
('014. Zones',
'Search by',
"""
if (SIDSTYPE in [CGK.Zone_ts]):
    RESULT=True
""",
"""
Search by
All <b>Zone_t</b> nodes.
""", False, False),
# -----------------------------------------------------------------------------
('015. Zones Structured',
'Search by',
"""
if (SIDSTYPE in [CGK.Zone_ts]):
    t=CGU.hasChildName(NODE,CGK.ZoneType_s)
    if (t is None or CGU.stringValueMatches(t,CGK.Structured_s)):
      RESULT=True
""",
"""
Search by
All <b>Zone_t</b> with Structured <b>ZoneType</b> nodes.
""", False, False),
# -----------------------------------------------------------------------------
('016. Zones Unstructured',
'Search by',
"""
if (SIDSTYPE in [CGK.Zone_ts]):
    t=CGU.hasChildName(NODE,CGK.ZoneType_s)
    if (t is not None and CGU.stringValueMatches(t,CGK.Unstructured_s)):
      RESULT=True
""",
"""
Search by
All <b>Zone_t</b> with Unstructured <b>ZoneType</b> nodes.
""", False, False),
# -----------------------------------------------------------------------------
('017. BCs',
'Search by',
"""
if (SIDSTYPE in [CGK.BC_ts]):
    RESULT=True
""",
"""
Search by
All <b>BC_t</b> nodes.
""", False, False),

# --- Replace
# -----------------------------------------------------------------------------
('050. Valued UserDefinedData',
'Replace',
"""
if (     (SIDSTYPE==CGK.UserDefinedData_ts)
     and (CGU.getValueDataType(NODE)!=CGK.MT)):
  NODE[3]=CGK.DataArray_ts
  RESULT=True
""",
"""
Replace
Valued UserDefinedData
 
Search all <b>UserDefinedData_t</b> nodes with a non-<b>MT</b> data type
and replace them as <b>DataArray_t</b>.""", False, False),

('051. Substitute Zone name',
'Replace',
"""
l1=len(ARGS[0])
if ((SIDSTYPE==CGK.Zone_ts) and (NAME[:l1]==ARGS[0])):
  NODE[0]=ARGS[1]+NODE[0][l1:]
  RESULT=True

if (CGU.getValueDataType(NODE)==CGK.C1):
  v=VALUE.tostring().decode('ascii')
  if (v[:l1]==ARGS[0]):
    v=ARGS[1]+v[l1:]
    NODE[1]=CGU.setStringAsArray(v)
    RESULT=True
""",
"""
<h1>Replace</h1>
<h2>Substitute Zone name</h2>
<p>
Search all <b>Zone_t</b> nodes with a name pattern, then rename the
zone with the substitution pattern. Any other reference in the tree,
as a connectivity value for example, is subsitued as well.
<p>
Argument is a tuple with the first pattern to find and
the second as the subsitution pattern. For example:
<pre>
('domain.','zone#')
</pre>
""", True, True),

# -----------------------------------------------------------------------------
('052. FamilySpecified BC type rewrite',
'Replace',
"""
if ((SIDSTYPE in [CGK.FamilyName_ts, CGK.AdditionalFamilyName_ts]) and
    (VALUE.tostring().decode('ascii')==ARGS[0]) and (PARENT[3]==CGK.BC_ts)):
    PARENT[1]=CGU.setStringAsArray(CGK.FamilySpecified_s)
    RESULT=True
""",
"""
<h1>Replace</h1>
<h2>FamilySpecified BC type rewrite</h2>
<p>
Search all <b>FamilyName BC</b> nodes with the arg string (plain).<br>
The string arg should be a valid Python string such as:<br>

'BLADE(1)'<br>
'Surface ext 1A'<br>
'Left/Wing/Flap'<br>

<p>
Once found, the parent <b>BC_t</b> value is forced to <b>FamilySpecified</b>

""", True, True),

# --- Find Elements_t
('020. Elements',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=True
""",
"""Find all <b>Elements_t</b> nodes """, False, False),

('021. Elements QUAD',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=VALUE[0] in (CGK.QUAD_4, CGK.QUAD_8, CGK.QUAD_9)
""",
"""Find all <b>Elements_t</b> nodes of type <b>QUAD</b>""", False, False),

('022. Elements TRI',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=VALUE[0] in (CGK.TRI_3, CGK.TRI_6)
""",
"""Find all <b>Elements_t</b> nodes of type <b>TRI</b>""", False, False),

('023. Elements NGON',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=VALUE[0] in (CGK.NGON_n,)
""",
"""Find all <b>Elements_t</b> nodes of type <b>NGON_n</b>""", False, False),

('024. Elements HEXA',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=VALUE[0] in (CGK.HEXA_8, CGK.HEXA_20, CGK.HEXA_27)
""",
"""Find all <b>Elements_t</b> nodes of type <b>HEXA</b>""", False, False),

('025. Elements TETRA',
'Find Elements_t',
"""if (SIDSTYPE==CGK.Elements_ts):
  RESULT=VALUE[0] in (CGK.TETRA_4, CGK.TETRA_10)
""",
"""Find all <b>Elements_t</b> nodes of type <b>TETRA</b>""", False, False),

# --- External Tools
('030. Create Cartesian Zone',
'External Tools',
"""
if (SIDSTYPE==CGK.CGNSTree_ts):
    import Generator.PyTree as G
    z=G.cart((0.,0.,0.), (0.1,0.1,0.2), (10,11,12))
    b=None
    base='BASE'
    if (len(ARGS)>0):
      base=ARGS[0]
      b=CGU.hasChildName(NODE,base)
    if (b is None):
      base=CGU.checkUniqueChildName(NODE,base)
      b=CGL.newCGNSBase(NODE,base,3,3)
    CGU.addChild(b,z)
""",
"""Example of Cartesian zone creation using Cassiopee.
The first argument is the base name, if ommitted a name is generated.""",
True, True),
('031. Bounding boxes',
'External Tools',
"""
if (SIDSTYPE==CGK.Zone_ts):
    import Generator as G
    RESULT=G.bbox(NODE)
""",
"""Example of Bounding box computation using Cassiopee"""),
('100. .Solver#Compute children',
'Edit filters',
"""
if (PARENT[0]=='.Solver#Compute'):
    RESULT=PATH
""",
"""Selects all children nodes of the .Solver#Compute elsA userdefined node""",
False, False),
('101. ReferenceState children',
'Edit filters',
"""
if (PARENT[0]=='ReferenceState'):
    RESULT=PATH
""",
"""Selects all children nodes of the ReferenceState node""",
False, False),
('102. .Solver#Param children',
'Edit filters',
"""
if (PARENT[0]=='.Solver#Param'):
   RESULT=PATH
""",
"""Selects all children nodes of the .Solver#Param elsA userdefined node""",
False, False),
    
    ]

    # -----------------------------------------------------------------
    @classmethod
    def _setOption(cls, name, value):
        setattr(cls, name, value)

    @classmethod
    def _writeFile(cls, tag, name, udata, filename, prefix=""):
        gdate = strftime("%Y-%m-%d %H:%M:%S", gmtime())
        s = u"""# %s - %s - Generated %s\n# coding: utf-8\n%s\nimport PyQt5\n%s=""" % \
            (cls._ToolName, tag, gdate, prefix, name)
        if isinstance(udata, dict):
            s += u"""{\n"""
            for k in udata:
                #          print 'K',k,'V',udata[k],'T',type(udata[k])
                if k[0] != '_':
                    val = '%s' % str(udata[k])
                    if isinstance(udata[k], str):
                        val = u'u"""%s"""' % repr(udata[k]).lstrip("u'").lstrip("'").rstrip("'") # Error
                    elif isinstance(udata[k], bytes):
                        val = u'u"""%s"""' % repr(udata[k].decode('utf-8')).lstrip("u'").lstrip("'").rstrip("'")
                    if not isinstance(k, str):
                        uk = u"u'%s'" % repr(k.decode('utf-8')).lstrip("u'").lstrip("'").rstrip("'")
                    else:
                        uk = u"u'%s'" % repr(k).lstrip("u'").lstrip("'").rstrip("'")
                    s += """%s:%s,\n""" % (uk, val)
            s += u"""}\n\n# --- last line\n"""
        elif isinstance(udata, list):
            s += u"""[\n"""
            for k in udata:
                s += u"""%s,\n""" % (k)
            s += u"""]\n\n# --- last line\n"""
        cls._crpath(filename)
        with open(filename, 'w+') as f:
            f.write(s)


    @classmethod
    def _readFile(cls, name, filename):
        dpath = tempfile.mkdtemp()
        if not op.exists(filename):
            return None
        try:
            copyOneFile(filename, '%s/%s.py' % (dpath, name))
        except IOError:
            removeSubDirAndFiles(dpath)
            return None
        sprev = sys.path
        sys.path = [dpath] + sys.path
        try:
            fp, pathname, description = imp.find_module(name)
        except IndexError:  # ImportError:
            return None
        try:
            mod = imp.load_module(name, fp, pathname, description)
        finally:
            if fp:
                fp.close()
        removeSubDirAndFiles(dpath)
        sys.path = sprev
        return mod

    @classmethod
    def _crpath(cls, path):
        p = op.dirname(path)
        if op.exists(p):
            return True
        os.makedirs(p)

    @classmethod
    def _trpath(cls, path):
        return op.normpath(op.expanduser(op.expandvars(path)))

    @classmethod
    def _writeHistory(cls, control):
        filename = cls._trpath(cls._HistoryFileName)
        cls._writeFile('History', 'history', control._history, filename)

    @classmethod
    def _readHistory(cls, control):
        filename = cls._trpath(cls._HistoryFileName)
        m = cls._readFile('history', filename)
        if m is None:
            return None
        try:
            return m.history
        except:
            return None

    @classmethod
    def _writeOptions(cls, control):
        filename = cls._trpath(cls._OptionsFileName)
        cls._writeFile('User options', 'options', control._options, filename)

    @classmethod
    def _readOptions(cls, control):
        filename = cls._trpath(cls._OptionsFileName)
        m = cls._readFile('options', filename)
        if m is None:
            return None
        try:
            return m.options
        except:
            return None

    @classmethod
    def _writeQueries(cls, control, q):
        filename = cls._trpath(op.join(cls.QueriesDirectory, cls._QueriesDefaultFile))
        cls._writeFile('User queries', 'queries', q, filename, cls.Q_FILE_PRE)

    @classmethod
    def _readQueries(cls, control):
        filename = cls._trpath(op.join(cls.QueriesDirectory, cls._QueriesDefaultFile))
        m = cls._readFile('queries', filename)
        if m is None:
            return None
        try:
            return m.queries
        except:
            return None

    @classmethod
    def _writeFunctions(cls, control, f):
        filename = cls._trpath(op.join(cls.FunctionsDirectory, cls._FunctionsDefaultFile))
        cls._writeFile('User functions', 'functions', f, filename, cls.Q_FILE_PRE)

    @classmethod
    def _readFunctions(cls, control):
        filename = cls._trpath(op.join(cls.FunctionsDirectory, cls._FunctionsDefaultFile))
        m = cls._readFile('functions', filename)
        if m is None:
            return None
        try:
            return m.Q7UserFunction
        except:
            return None

    def __init__(self):
        pass

    def __getitem__(self, name):
        if name[0] != '_':
            return Q7OptionContext.__dict__[name]
        return None

    def __setitem__(self, name, value):
        if name[0] != '_':
            setattr(Q7OptionContext, name, value)
        return None

    def __iter__(self):
        for o in dir(self):
            if o[0] != '_':
                yield o

    def _nextName(self):
        for o in dir(self):
            if o[0] != '_':
                yield o

# -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (str, bytes, range, dict)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

from qtpy.QtCore import Qt, QEvent
from qtpy.QtWidgets import QComboBox, QListWidget

# -----------------------------------------------------------------
class Q7ComboBox(QComboBox):
    def __init__(self, arg):
        QComboBox.__init__(self, arg)
        self.actorlist = QListWidget()
        self.setModel(self.actorlist.model())
        self.setView(self.actorlist)
        self.view().installEventFilter(self)
        self.parent = None

    def setParent(self, parent):
        self.parent = parent

    def eventFilter(self, o, e):
        if e.type() == QEvent.KeyPress:
            # kmod = e.modifiers()
            kval = e.key()
            if kval in [Qt.Key_Z]:
                path = self.actorlist.currentItem().text()
                actor = self.parent.findPathObject(path)
                self.parent.changeCurrentActor([path, actor])
                return True
            if kval in [Qt.Key_H]:
                path = self.actorlist.currentItem().text()
                actor = self.parent.findPathObject(path)
                self.parent.changeCurrentActor([path, actor])
                self.parent.hideActor(None)
                return True
        return QComboBox.eventFilter(self, o, e)

    def keyPressEvent(self, event):
        kmod = event.modifiers()
        kval = event.key()
        print(kmod, kval)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
import sys
import os
import os.path
import fnmatch
import imp
import CGNS.VAL.grammars.etablesids as STB

PROFILENAME = 'grammars'


#  -------------------------------------------------------------------------
def readProfile():
    try:
        hdir = os.environ['HOME']
    except:
        return {}
    pdir = '%s%s.CGNS.NAV' % (hdir, os.path.sep)
    if not os.path.exists(pdir):
        return {}
    sys.path.append(pdir)
    fp, pth, des = imp.find_module('grammars')
    try:
        mod = imp.load_module('grammars', fp, pth, des)
    finally:
        if fp:
            fp.close()
    return mod.Grammars


#  -------------------------------------------------------------------------
def findAllUserGrammars(verbose=False):
    kdict = {}
    for pth in [p for p in sys.path if p != '']:
        if verbose:
            print('### scanning', pth)
        try:
            for pthroot, dirs, files in os.walk(pth):
                for fn in files:
                    if (fnmatch.fnmatch(fn, 'CGNS_VAL_USER_*.py') or
                            fnmatch.fnmatch(fn, 'CGNS_VAL_USER_*.so')):
                        gkey = fn[14:-3]
                        if (gkey in kdict):
                            if (verbose):
                                print('### * found grammar:', )
                                print(gkey, 'already found, ignore this one')
                        else:
                            if (verbose):
                                print('### * found grammar:', gkey)
                            kdict[fn[14:-3]] = pthroot
                        if (verbose):
                            print('### * found in :', pthroot)
                            if (pthroot not in sys.path):
                                print('### * previous path is NOT in PYTHONPATH')
                            else:
                                print('### * previous path already is in PYTHONPATH')

        except OSError:
            pass
    return kdict


#  -------------------------------------------------------------------------
def findOneUserGrammar(tag, verbose=False):
    kdict = {}
    found = False
    for pth in sys.path:
        if verbose:
            print('### scanning', pth)
        try:
            for pthroot, dirs, files in os.walk(pth):
                for fn in files:
                    if fnmatch.fnmatch(fn, 'CGNS_VAL_USER_%s.py' % tag):
                        kdict[fn[14:-3]] = pthroot
                        found = True
                        break
                if found:
                    break
        except OSError:
            pass
        if found:
            break
    return kdict


#  -------------------------------------------------------------------------
def importUserGrammars(key, recurse=False, verbose=False):
    mod = None
    modname = 'CGNS_VAL_USER_%s' % key
    ipath = os.path.split(STB.__file__)[0]
    sys.path.append(ipath)

    if (verbose):
        print('### Looking for grammar [%s]' % key)
    try:
        tp = imp.find_module(modname)
    except ImportError:
        if (verbose):
            print('### Error: grammar [%s] not found' % key)
        if (recurse):
            dk = findOneUserGrammar(key)
            if (key in dk):
                sys.path.append(dk[key])
                if (verbose):
                    print('### Warning: not in search path [%s]' % dk[key])
                try:
                    tp = imp.find_module(modname)
                except ImportError:
                    return None
            else:
                return None
        else:
            return None
    try:
        fp = tp[0]
        if (tp[2][2] != imp.C_EXTENSION):
            mod = imp.load_module(modname, *tp)
        else:
            # print '### CGNS.VAL [info]: Module info',tp
            mod = imp.load_dynamic(modname, tp[1], tp[0])
    except:
        pass
    finally:
        if fp:
            fp.close()
    return mod


#  -------------------------------------------------------------------------
def locateGrammars():
    glist = []
    for k in ['SIDS', 'elsA']:
        mod = importUserGrammars(k)
        if mod is not None:
            glist.append((k, mod.__file__))
    return glist

# ---



# -*- coding: utf-8 -*-
#
#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from __future__ import print_function
from builtins import (str, bytes, range, dict)

from CGNS.APP.lib.queries import asQuery
from CGNS.APP.lib.queries import runQuery


@asQuery
def nodeNameSearch(C):
    """
    Search by
    Node name

    Search all nodes with the exact NAME as argument.

    The argument name need not to be a tuple or to have quotes,
    all the following values are ok and would match the NAME <i>ZoneType</i>:

    ZoneType
    'ZoneType'
    ('ZoneType',)
    """
    if C.NAME == C.ARGS[0]:
        return C.PATH


@asQuery
def nodeTokenSearch(C):
    """
    Search by
    Node token

    Search all nodes with NAME as a token in the PATH

    The argument name need not to be a tuple or to have quotes,
    all the following values are ok and would match the NAME <i>ZoneBC</i>:

    ZoneBC
    'ZoneBC'
    ('ZoneBC',)
    """
    if C.ARGS[0] in C.PATH:
        return C.PATH


FILE = 'HYB/vbv-part32_comp_period_links_dom_32_SUB.hdf'
ARGS = ['GridLocation']

# -----------------------------------------------------------------
import CGNS.MAP as CGM

(t, l, p) = CGM.load(FILE)
print(runQuery(t, l, p, nodeNameSearch, ['GridLocation']))
print(runQuery(t, l, p, nodeTokenSearch, ['Zone']))



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from __future__ import unicode_literals
from builtins import (str, bytes, range, dict)

def asQuery(f):
    def prepostquery(node, parent, tree, links, skips, path, args, selected):
        global CGK
        global CGU
        global CGL
        global numpy
        global NODE
        global PARENT
        global NAME
        global VALUE
        global CGNSTYPE
        global CHILDREN
        global TREE
        global PATH
        global LINKS
        global SKIPS
        global RESULT
        global USER
        global SELECTED
        global RESULT_LIST

        import CGNS.PAT.cgnskeywords as CGK
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnslib as CGL
        import CGNS.MAP as CGM
        import numpy

        QueryNoException = True

        NODE = node
        PARENT = parent
        NAME = node[0]
        VALUE = node[1]
        CGNSTYPE = node[3]
        CHILDREN = node[2]
        TREE = tree
        PATH = path
        LINKS = links
        SKIPS = skips
        USER = args
        SELECTED = selected
        RESULT_LIST = [False]

        if QueryNoException:
            RESULT = f()
        else:
            try:
                RESULT = f()
                if RESULT not in [True, False]:
                    RESULT = False
                RESULT_LIST[0] = RESULT
            except Exception:
                RESULT_LIST[0] = False

        return RESULT

    return prepostquery


# -----------------------------------------------------------------
def parseAndSelect(tree, node, parent, links, skips, path, query, args, selected,
                   result):
    path = path + '/' + node[0]
    Q = query(node, parent, tree, links, skips, path, args, selected)
    R = []
    if Q:
        if result:
            R = [Q]
        else:
            R = [path]
    for C in node[2]:
        R += parseAndSelect(tree, C, node, links, skips, path, query, args, selected,
                            result)
    return R


# -----------------------------------------------------------------
def runQuery(tree, links, paths, query, args, selected=None, mode=True):
    """Recursively applies a function on all nodes of a tree, breadth-first
    parse, results returned in a list (breadth-first order of the list).

    tree,links,paths: the three restults returned by the CGNS.MAP.load call
    query: the query funciton as a name (string) or a callable object (such as
    a function)
    args: a tuple of args to pass to the query (for example a node name to
    look for)
    selected: the RETURNED list of values
    mode: True (defaut) a boolean list is returned for every node in the tree,
    False a list of the (True) paths is returned
    
    The match between results and paths can be performed with help of the
    breadth-first paths order function in CGNS.PAT
    """
    if selected is None:
        selected = []
    v = None
    try:
        if args:
            v = eval(args)
        if (v is not None) and not isinstance(v, tuple):
            v = (v,)
    except NameError:
        v = (str(args),)
    except:
        pass
    _args = v
    if isinstance(query, (str, bytes)):
        query = eval(query)
    result = parseAndSelect(tree, tree, [None, None, [], None], links, paths, '',
                            query, _args, selected, mode)
    return result



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System - 
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
#
from CGNS.APP.lib import mergeTrees
import CGNS.MAP as CGM


def treeLinkZGC(T1):
    pass


def fileLinkZGC(*files):
    l = []
    for f in files:
        l += [CGM.load(f)[0]]




import CGNS.PAT.cgnsutils    as CGU
import CGNS.PAT.cgnskeywords as CGK

def getBCbyFamily(T,family):
    """Finds all BC that belongs to a family.
    Works with more than one base.
    Returns a list of path, you have to use getNodeByPath to retrieve
    the actual node."""
    result=[]
    fpath=[CGK.CGNSTree_ts,CGK.CGNSBase_t,CGK.Zone_t,CGK.ZoneBC_t,
           CGK.BC_t,CGK.FamilyName_t]
    flist=CGU.getAllNodesByTypeOrNameList(T,fpath)
    for famnamepath in flist:
        fnode=CGU.getNodeByPath(T,famnamepath)
        if (CGU.stringValueMatches(fnode,family)):
            bcpath=CGU.getAncestor(famnamepath)
            result.append(bcpath)
    return result
